from enum import Enum
from json import dumps, load
from typing import Type

from hses_genesis.utils.constants import MAX_HOSTS_PER_SWITCH_KEY, TERMINAL_WIDTH, LAYER_DEFINITIONS_KEY, STRUCTURE_KEY, SECURITY_KEY, COMMUNICATION_KEY, TOPOLOGY_KEY, ITERATIONS_KEY, SUBNET_DESCENDANTS_KEY, SWITCH_COUNT_KEY, HOST_TYPES_KEY, ANOMALY_COUNT_KEY, STATEFUL_PERCENTAGE_KEY, TRAFFIC_PROFILE_KEY, UPPER_CONNECTION_BOUND_KEY
from hses_genesis.utils.enum_objects import EDeviceRole, ESubnetTopologyStructure
from hses_genesis.utils.functions import load_resource

class EResponse(Enum):
    ANSWER = 0
    HELP = 1
    INTERRUPT = 2
    INVALID = 3
    NONE = 4
    SKIP = 5
    BACK = 6

class ConfigurationGenerator():
    def __init__(self):
        self.config = {
            SECURITY_KEY : dict(),
            COMMUNICATION_KEY : dict(),
            TOPOLOGY_KEY : {
                LAYER_DEFINITIONS_KEY : []
            }
        }

        hint_location = load_resource('configurations', 'configuration_hints.json')
        with open(hint_location, 'r') as f:
            self.parameter_info = dict(load(f))

    def sloppy_answer(self, question : str, possible_values : list[str]):
        while True:
            answer = input(question)
            matches = [x for x in possible_values if x.lower().startswith(answer.lower())]
            if len(matches) != 1:
                print('Your input did not match any of the allowed options. Please try again.')
                continue
            
            return matches[0]

    def interpret_answer(self, answer, answer_type : Type, possible_values = None):
        if not answer:
            return EResponse.NONE, None
        
        if answer == 'h' or answer == 'help':
            return EResponse.HELP, None
        
        if answer == 'b':
            return EResponse.BACK, None
        
        if answer == 's':
            return EResponse.SKIP, None
        
        if answer == 'x' or answer == 'cancel' or answer == 'exit':
            return EResponse.INTERRUPT, None
        
        if (possible_values and answer not in possible_values):
            return EResponse.INVALID, f'Your response did not match any of the allowed options ({str(possible_values)}).'
        
        try:
            return EResponse.ANSWER, answer_type(answer)
        except:
            return EResponse.INVALID, f'Your response is not of type {answer_type}.'
        
    def parameter_question(self, parameter, answer_type : Type, default_value = None, possible_values = None, allow_back = False, allow_skip = False, help_key = None):
        if default_value == None:
            default_value = dict(self.parameter_info.get(parameter, dict())).get('default_value', None)

        if possible_values == None:
            possible_values = dict(self.parameter_info.get(parameter, dict())).get('valid_options', None)
        
        question = f'Please provide your {parameter} value'
        if possible_values != None:
            question += f' {possible_values}'
        if allow_skip:
            question += ', skip it (s)'
        if allow_back:
            question += ', go back (b)'
        question += f', ask for help (h) or exit (x)' + (f' (default value: {default_value}): ' if default_value != None else ': ')
        
        while True:
            answer = input(question)
            response_type, value = self.interpret_answer(answer, answer_type, possible_values)
            if response_type == EResponse.NONE:
                if default_value != None:
                    return EResponse.ANSWER, answer_type(default_value)
                else:
                    response_type = EResponse.INVALID

            if response_type == EResponse.BACK:
                if not allow_back:
                    response_type = EResponse.INVALID
                else:
                    return EResponse.BACK, None
                
            if response_type == EResponse.SKIP:
                if not allow_skip:
                    response_type = EResponse.INVALID
                else:
                    return EResponse.SKIP, answer_type(default_value)
            
            if response_type in [EResponse.ANSWER, EResponse.INTERRUPT]:
                return response_type, value
            
            if response_type == EResponse.HELP:
                help_parameter = parameter if help_key == None else help_key
                print(f'"{help_parameter}"', self.parameter_info.get(help_parameter, {'info' : f'Sorry, could not find infos for parameter "{parameter}".'})['info'])

            if response_type == EResponse.INVALID:
                print(value, 'Please try again.')
                continue

    def print_config_info(self, key : str = None, to_print : dict = None, no_header = False):
        if not no_header:
            print('-' * TERMINAL_WIDTH)
            if not key:
                print(f'Your current configuration:')
            else:
                print(f'Your current {key.lower()} configuration:')
            print('-' * TERMINAL_WIDTH)
        if to_print == None:
            to_print = self.config
        print(dumps(to_print, indent=4))
        print('-' * TERMINAL_WIDTH)

    def edit_config(self):
        while True:
            self.print_config_info()

            value = self.ask_user('What do you want to do? Configure (t) topology, (c) communication, (s) security, or (x) save your current configuration and start generation: ', possible_answers=['t', 'c', 's', 'x'])
            if value == 't':
                self.edit_topology()
            if value == 'c':
                self.edit_communication()
            if value == 's':
                self.edit_security()
            if  value == 'x':
                if len(self.config[TOPOLOGY_KEY][LAYER_DEFINITIONS_KEY]) == 0:
                    print('=' * TERMINAL_WIDTH)
                    print('WARNING: Your configuration is invalid! At least one layer definition is required! Please configure at least one layer definition:')
                    print('=' * TERMINAL_WIDTH)
                    self.edit_topology('a')
                else:
                    break
        return self.config

    def setup_segment(self, contained_keys : dict[str, dict], allow_pagination = True, help_key = None):
        segment = {
            key : value['default_value'] for key, value in contained_keys.items() if 'default_value' in value.keys() 
        }
        items = list(contained_keys.items())
        i = 0
        while i <= len(contained_keys) - 1:
            key, infos = items[i]
            allow_back = allow_pagination and i > 0
            while True:
                response_type, value = self.parameter_question(key, infos['typing'], default_value = segment.get(key, None), allow_skip=allow_pagination, allow_back=allow_back, help_key=help_key)
                if response_type == EResponse.SKIP and allow_pagination:
                    break

                if response_type == EResponse.BACK and allow_back:
                    i -= 2
                    break

                if response_type == EResponse.INTERRUPT:
                    return response_type, None

                if response_type == EResponse.ANSWER:
                    segment[key] = value
                    break
            i += 1
        return response_type, segment

    def edit_security(self):
        self.print_config_info(SECURITY_KEY, self.config[SECURITY_KEY])

        response, segment = self.setup_segment({
            x: {
                'typing' : int,
                'default_value' : None
            } for x in [ITERATIONS_KEY, ANOMALY_COUNT_KEY, STATEFUL_PERCENTAGE_KEY]
        })
        if response == EResponse.ANSWER:
            self.config[SECURITY_KEY] = segment        

    def edit_communication(self):
        self.print_config_info(COMMUNICATION_KEY, self.config[COMMUNICATION_KEY])

        response, segment = self.setup_segment({
            ITERATIONS_KEY: {
                'typing' : int,
                'default_value' : None
            },
            TRAFFIC_PROFILE_KEY: {
                'typing' : str,
                'default_value' : None
            },
            UPPER_CONNECTION_BOUND_KEY: {
                'typing' : int,
                'default_value' : None
            },
        })
        if response == EResponse.ANSWER:
            self.config[COMMUNICATION_KEY] = segment        

    def ask_user(self, question, possible_answers = None, answer_type : Type = str):
        while True:
            answer = input(question).lower()
            if possible_answers and answer not in possible_answers:
                print('You did not provide a valid option. Please try again.')
                continue
            
            try:
                return answer_type(answer)
            except:
                print(f'Your answer is not of type {answer_type}. Please try again.')        

    def get_host_types(self, switch_count : int, hosts_per_switch : int, default_values : dict = None):
        def fixed_host_count(values : dict[str, int]):
            return sum(v for v in values.values() if v > 0)

        def validity_check(values : dict[str, int]):
            max_host_count = switch_count * hosts_per_switch
            remaining_hosts = max_host_count - fixed_host_count(values)
            return remaining_hosts >= 0

        if default_values == None:
            default_values = {key.name : 0 for key in EDeviceRole.configurables()}

        self.print_config_info(HOST_TYPES_KEY, default_values)

        while True:
            
            value_info = {key.name : {
                'typing' : int,
                'default_value' : default_values.get(key.name, 0)
            } for key in EDeviceRole.configurables()}

            response, default_values = self.setup_segment(contained_keys=value_info, help_key=HOST_TYPES_KEY)
            if response == EResponse.INTERRUPT:
                break

            if validity_check(default_values):
                return response, default_values
            else:
                print('=' * TERMINAL_WIDTH)
                print(f'WARNING: Your host_types configuration is invalid! You defined too many fixed devices (hosts_per_switch ({hosts_per_switch}) * switch_count ({switch_count}) < {fixed_host_count(default_values)}). Please adjust your configuration.')
                print('=' * TERMINAL_WIDTH)
                self.print_config_info(HOST_TYPES_KEY, default_values, no_header=True)
        
    def get_structure_distribution(self, default_values = None):
        if default_values == None:
            default_values = {key.name : 0 for key in ESubnetTopologyStructure}
        
        self.print_config_info(STRUCTURE_KEY, default_values)

        default_values = {
            key.name : {
            'typing' : int,
            'default_value' : default_values.get(key.name, 0)
            } for key in ESubnetTopologyStructure
        }

        return self.setup_segment(default_values, help_key=STRUCTURE_KEY)

    def edit_layer_definition(self, layer_index = -1):
        if layer_index < 0:
            print(f'-- Adding a new layer --')
            input_definition = dict()
        else:
            print(f'-- Editing layer {layer_index} --')
            input_definition = self.config[TOPOLOGY_KEY][LAYER_DEFINITIONS_KEY][int(layer_index)]
        
        response_type, layer_definition = self.setup_segment({
            x: {
                'typing' : int,
                'default_value' : input_definition.get(x, None)
            } for x in [SUBNET_DESCENDANTS_KEY, SWITCH_COUNT_KEY, MAX_HOSTS_PER_SWITCH_KEY]
        })

        if response_type == EResponse.INTERRUPT:
            return
        
        response_type, host_types = self.get_host_types(layer_definition.get(SWITCH_COUNT_KEY), layer_definition.get(MAX_HOSTS_PER_SWITCH_KEY), input_definition.get(HOST_TYPES_KEY, None))
        
        if response_type == EResponse.INTERRUPT:
            return
        
        layer_definition[HOST_TYPES_KEY] = host_types

        response_type, structure = self.get_structure_distribution(input_definition.get(STRUCTURE_KEY, None))
        
        if response_type == EResponse.INTERRUPT:
            return
        
        layer_definition[STRUCTURE_KEY] = structure

        if layer_index < 0:
            self.config[TOPOLOGY_KEY][LAYER_DEFINITIONS_KEY].append(layer_definition)
        else:
            self.config[TOPOLOGY_KEY][LAYER_DEFINITIONS_KEY][layer_index] = layer_definition

        
    def edit_topology(self, intro_step = None):
        def handle_response(answer : str):
            if answer in ['', 'a']:
                self.edit_layer_definition()
            
            if answer == 'e':
                layer_index = 0
                if len(layer_definitions) > 1:
                    layer_index = self.ask_user(f'Which layer do you want to edit? Please provide its index [0..{len(layer_definitions) - 1}]: ', answer_type=int, possible_answers=map(lambda x: str(x), range(0, len(layer_definitions))))
                self.edit_layer_definition(layer_index)
            
            if answer == 'r':
                layer_index = self.ask_user(f'Which layer do you want to remove? Please provide its index [0..{len(layer_definitions) - 1}]: ', answer_type=int, possible_answers=map(lambda x: str(x), range(0, len(layer_definitions))))
                print(f'You aim to remove layer {layer_index}:')
                print('-' * TERMINAL_WIDTH)
                print(dumps(layer_definitions[layer_index]))
                print('-' * TERMINAL_WIDTH)
                answer = self.ask_user('Are you shure you want to remove this layer (y - yes/n - no)?', ['y', 'n'])
                if answer == 'y':
                    self.config[TOPOLOGY_KEY][LAYER_DEFINITIONS_KEY].pop(layer_index)

        if intro_step:
            handle_response(intro_step)

        while True:
            self.print_config_info(TOPOLOGY_KEY, self.config[TOPOLOGY_KEY])
            
            layer_definitions = self.config[TOPOLOGY_KEY][LAYER_DEFINITIONS_KEY]
            question = 'What do you want to do? (a - add layer,'
            possible_answers = ['', 'a', 'x']
            if len(layer_definitions) > 0:
                question += ' r - remove layer, e - edit layer,'
                possible_answers += ['r', 'e']
            question += ' x - exit (default value a): '
            answer = self.ask_user(question, possible_answers=possible_answers)

            if answer == 'x':
                return
            
            handle_response(answer)


from enum import Enum
from types import SimpleNamespace

class EGenerationSteps(Enum):
    TOPOLOGY = 0
    COMMUNICATION_RELATIONS = 1
    NETWORK_CONFIGURATIONS = 2

class ENetworkLayer(Enum):
    CONNECTIVITY = 0
    AGGREGATED_CONTROL = 1
    PROCESS = 2

    @staticmethod
    def from_str(input : str):
        for v in ENetworkLayer:
            if v.name.lower().startswith(input.lower()):
                return v
            
        raise Exception(f'No matching network layer found for {input}')
    
    def from_value(input : int):
        for v in ENetworkLayer:
            if v.value == input:
                return v
            
        raise Exception(f'No matching network layer found for {input}')

class EPacketDecision(Enum):
    DROP = 0
    ACCEPT = 1
    
    def negate(self):
        if self == EPacketDecision.ACCEPT:
            return EPacketDecision.DROP
        return EPacketDecision.ACCEPT
    
    @staticmethod
    def from_str(value):
        for v in EPacketDecision:
            if str(v) == value or v.name == value:
                return v
        raise Exception(f'Invalid packet decision: {value}')

class ESubnetTopologyStructure(Enum):
    STAR = 0
    RING = 1
    LINE = 2
    MESH = 3
    
    @staticmethod
    def from_str(value : str):
        for v in ESubnetTopologyStructure:
            if v.name.lower() == value.lower() or v.name.lower().startswith(value.lower()):
                return v
        raise Exception(f'Invalid Subnet Type: {value}. Use one of {list(ESubnetTopologyStructure.__members__.values())}.')
    
    @staticmethod
    def redundant_types():
        return [ESubnetTopologyStructure.RING, ESubnetTopologyStructure.MESH]
        

class EState(Enum):
    NONE = 0
    NEW = 1
    ESTABLISHED = 2
    RELATED = 3
    INVALID = 4
    UNTRACKED = 5

    @staticmethod
    def from_value(input : int):
        for v in EState:
            if v.value == input:
                return v
            
        raise Exception(f'NO VALID STATE ({input}). Use one of {list(EState.__members__.values())}')
    

class EService(Enum):
    SSH = SimpleNamespace(**{'protocols' : ['tcp'], 'ports' : [22], 'packet_size_range' : [512,1500], 'priority' : 1, 'dei' : 0})
    HTTP = SimpleNamespace(**{'protocols' : ['tcp'], 'ports' : [80], 'packet_size_range' : [64,1500], 'priority' : 1, 'dei' : 0})
    HTTPS = SimpleNamespace(**{'protocols' : ['tcp'], 'ports' : [443], 'packet_size_range' : [64,1500], 'priority' : 1, 'dei' : 0})
    OPC_UA = SimpleNamespace(**{'protocols' : ['tcp'], 'ports' : [4840], 'packet_size_range' : [64,800], 'priority' : 1, 'dei' : 0})
    NETCONF = SimpleNamespace(**{'protocols' : ['tcp'], 'ports' : [830], 'packet_size_range' : [64,800], 'priority' : 1, 'dei' : 0})
    IP_Camera = SimpleNamespace(**{'protocols' : ['tcp'], 'ports' : [554], 'packet_size_range' : [1000,1500], 'priority' : 1, 'dei' : 0}) # rtsp uses mostly tcp --> no signature protocol number in ip header
    ModbusTCP = SimpleNamespace(**{'protocols' : ['tcp'], 'ports' : [502], 'packet_size_range' : [64,800], 'priority' : 1, 'dei' : 0})
    EthernetIP = SimpleNamespace(**{'protocols' : ['tcp', 'udp'], 'ports' : [44818, 2222], 'packet_size_range' : [64,100], 'priority' : 1, 'dei' : 0})
    EtherCAT = SimpleNamespace(**{'protocols' : ['udp'], 'ports' : [34980], 'packet_size_range' : [64,100], 'priority' : 1, 'dei' : 0})
    ProfiNet = SimpleNamespace(**{'protocols' : ['udp'], 'ports' : [34962,34963,34964,53247], 'packet_size_range' : [64,100], 'priority' : 1, 'dei' : 0})

    def __str__(self) -> str:
        return self.name
    
    @staticmethod
    def from_str(value):
        for v in EService:
            if v.name == value:
                return v
            
        raise Exception(f'Unknown Service passed: {value}.')

    def to_dict(self):
        protocols, ports, _, _ = self.value
        return {
            "service": self.name, 
            "protocols": protocols, 
            "ports": ports
        }

class EDeviceRole(Enum):
    SERVER = 'SV'
    ROUTER = 'R'
    SWITCH = 'SW'
    IT_END_DEVICE = 'I'
    CONTROLLER = 'C'
    OT_END_DEVICE = 'O'
    PORT = '#'
    
    @staticmethod
    def from_name(input : str):
        for v in EDeviceRole:
            if v.name.lower() == input.lower():
                return v
        raise Exception(f'No device role with name {input} supported.')
        
    @staticmethod
    def from_device_id(input : str):
        matches = [e for e in EDeviceRole if input.startswith(str(e.value))]
        if len(matches) != 1:
            raise Exception(f'Invalid Device ID: {input}')
        return matches[0]
    
    @staticmethod
    def configurables():
        return [EDeviceRole.SERVER, EDeviceRole.IT_END_DEVICE, EDeviceRole.OT_END_DEVICE, EDeviceRole.CONTROLLER]
    
    @staticmethod
    def from_configurables_id(value : str):
        for v in EDeviceRole.configurables():
            if value.startswith(v.name[0]):
                return v
        raise Exception(f'No configurable with name {value} supported.')
    
    def __str__(self) -> str:
        return self.name
    
    def possible_services(self) -> list[EService]:
        services = [EService.HTTP, EService.HTTPS]
        if self != EDeviceRole.OT_END_DEVICE:
            services.append(EService.SSH)

        if self != EDeviceRole.IT_END_DEVICE:
            services.append(EService.OPC_UA)

        if self not in [EDeviceRole.IT_END_DEVICE, EDeviceRole.OT_END_DEVICE]:
            services.append(EService.NETCONF)

        if self in [EDeviceRole.SERVER, EDeviceRole.IT_END_DEVICE]:
            services.append(EService.IP_Camera)

        if self in [EDeviceRole.SERVER, EDeviceRole.CONTROLLER, EDeviceRole.OT_END_DEVICE]:
            services.append(EService.ModbusTCP)
            services.append(EService.EthernetIP)
            services.append(EService.EtherCAT)
            services.append(EService.ProfiNet)

        return services

    def possible_service_states(self, service : EService):
        if self == EDeviceRole.SERVER:
            return [EState.NEW, EState.ESTABLISHED, EState.RELATED]
        
        if self in [EDeviceRole.ROUTER, EDeviceRole.SWITCH]:
            if service in [EService.HTTP, EService.HTTPS, EService.OPC_UA, EService.NETCONF]:
                return [EState.ESTABLISHED, EState.RELATED]
            return []
            
        if self == EDeviceRole.CONTROLLER:
            if service in [EService.HTTP, EService.HTTPS]:
                return [EState.ESTABLISHED, EState.RELATED]
            elif service != EService.IP_Camera:
                return [EState.NEW, EState.ESTABLISHED, EState.RELATED]
            
            return []
            
        if self == EDeviceRole.IT_END_DEVICE:
            if service in [EService.HTTP, EService.HTTPS, EService.SSH, EService.IP_Camera]:
                return [EState.NEW, EState.ESTABLISHED, EState.RELATED]
            return []

        if self == EDeviceRole.OT_END_DEVICE:
            if service in [EService.ModbusTCP, EService.EthernetIP, EService.EtherCAT, EService.ProfiNet]:
                return [EState.NEW, EState.ESTABLISHED, EState.RELATED]
            elif service in [EService.HTTP, EService.HTTPS, EService.OPC_UA]:
                return [EState.ESTABLISHED, EState.RELATED]

        return []
    
    @staticmethod
    def high_senders():
        return [EDeviceRole.IT_END_DEVICE, EDeviceRole.OT_END_DEVICE]
    
class ETrafficProfile(Enum):
    STRICT_ISOLATION = 0
    CONVERGED_NETWORKS = 1
    DISTRIBUTED_CONTROL = 2

    @staticmethod
    def from_str(input : str):
        input_value = input.upper().replace(' ', '_')
        for v in ETrafficProfile:
            if v.name == input_value:
                return v
        
        raise Exception(f'NO VALID TRAFFIC PROFILE ({input}). Use one of {list(ETrafficProfile.__members__.keys())}.')
    
    @staticmethod
    def from_value(input : int):
        for v in ETrafficProfile:
            if v.value == input:
                return v
        raise Exception(f'NO VALID TRAFFIC PROFILE ({input}). Use one of {list(ETrafficProfile.__members__.values())}.')
    
class EParameterKey(Enum):
    SRC = 's'
    DST = 'd'
    PROTOCOL = 'p'
    SRC_PORT = '-sport'
    DST_PORT = '-dport'

    @staticmethod
    def from_str(input : str):
        for v in EParameterKey:
            if v.name == input or v.value == input:
                return v
        raise Exception(f'No parameter key supported matching {input}. Use one of {list(EParameterKey.__members__.values())}.')

class EParameterType(Enum):
    IP = 0
    NUMBER = 1
    PROTOCOL = 2

    @staticmethod
    def from_parameter_key(parameter_key : EParameterKey):
        if parameter_key in [EParameterKey.SRC, EParameterKey.DST]:
            return EParameterType.IP
        elif parameter_key in [EParameterKey.PROTOCOL]:
            return EParameterType.PROTOCOL
        else:
            return EParameterType.NUMBER


from abc import ABC, abstractmethod
from json import dump, load
from random import randint
from hses_genesis.utils.constants import BIG_SEPARATOR, COMMUNICATION_KEY, GENESIS_PACKAGE_NAME, HOST_TYPES_KEY, ITERATION_SEPARATOR, ITERATIONS_KEY, LAYER_DEFINITIONS_KEY, MAX_HOSTS_PER_SWITCH_KEY, MEDIUM_SEPARATOR, SUBNET_DESCENDANTS_KEY, REPETITIONS_KEY, ANOMALY_COUNT_KEY, SECURITY_KEY, SMALL_SEPARATOR, STATEFUL_PERCENTAGE_KEY, STRUCTURE_KEY, SWITCH_COUNT_KEY, TAG_ENCAPSULATOR, TOPOLOGY_KEY, TRAFFIC_PROFILE_KEY, UNSET_INDICATOR, UPPER_CONNECTION_BOUND_KEY
from hses_genesis.utils.enum_objects import EDeviceRole, ESubnetTopologyStructure, ETrafficProfile, ENetworkLayer
from importlib.metadata import version
from re import match, search

def get_single_value_by_multiple_keys(value : dict, keys : list[str], default_value):
    for key in keys:
        if key in value.keys():
            return value.get(key)
        if key.lower() in value.keys():
            return value.get(key.lower())
    return default_value

class AGenesisTaggable(ABC):
    def __init__(self, use_latex_tag = False):
        super().__init__()
        self.COMPLETED = False
        self.use_latex_tag = use_latex_tag

    @abstractmethod
    def from_str(value, use_latex_tag = False):
        pass
    
    @abstractmethod
    def from_dict(value, use_latex_tag = False):
        pass

    @abstractmethod
    def to_run_str(self) -> str:
        pass

    def __str__(self):
        return self.to_run_str()

    @abstractmethod
    def to_run_dict_value(self):
        pass

    def to_total_str(self) -> str:
        return self.to_run_str()

    def to_total_dict_value(self):
        return self.to_run_dict_value()
    
    def iterate(self) -> bool:
        if self.COMPLETED:
            return False
        
        iteration_success = self.__apply_iteration_logic__()
        if not iteration_success:
            self.COMPLETED = True
        return iteration_success
    
    def __apply_iteration_logic__(self) -> bool:
        return False
    
    def reset(self):
        self.COMPLETED = False

class IterativeParameterValue(AGenesisTaggable):
    def __init__(self, start : int, end : int = None, step_size : int = 1, use_latex_tag = False):
        super().__init__(use_latex_tag)
        self.CURRENT, self.START = start, start
        self.END = end if end else start
        self.STEP_SIZE = step_size

    def to_run_str(self, default_value = None):
        if (not self.use_latex_tag) and default_value == self.CURRENT:
            return UNSET_INDICATOR
        return str(self.CURRENT)
    
    def to_total_str(self, default_value = None):
        if (not self.use_latex_tag) and self.START == self.END and default_value == self.START:
            return UNSET_INDICATOR
        if self.START == self.END:
            return str(self.CURRENT)
        return f'{self.START}{ITERATION_SEPARATOR}{self.END}{ITERATION_SEPARATOR}{self.STEP_SIZE}'
    
    def to_run_dict_value(self):
        return self.CURRENT

    def __int__(self):
        return self.CURRENT
    
    def __list__(self):
        if self.START == self.END:
            return [self.CURRENT]
        return list(range(self.CURRENT, self.END, self.STEP_SIZE))

    def to_total_dict_value(self):
        if self.START != self.END:
            return [self.START, self.END, self.STEP_SIZE]
        return self.CURRENT

    @staticmethod
    def from_dict(value, use_latex_tag = False, minimum = None, maximum = None):
        output = [int(x) for x in value] if isinstance(value, list) else [int(value)]
        if minimum:
            output = [max(minimum, x) for x in output]
        if maximum:
            output = [min(maximum, x) for x in output]

        if len(output) > 1:
            start, end, step_size = output
            return IterativeParameterValue(start, end, step_size, use_latex_tag=use_latex_tag)
        return IterativeParameterValue(output[0], use_latex_tag=use_latex_tag)
    
    @staticmethod
    def from_str(value : str, use_latex_tag = False, minimum = None, maximum = None):
        values = [int(x) for x in value.split(ITERATION_SEPARATOR)]
        return IterativeParameterValue.from_dict(values, use_latex_tag, minimum, maximum)
    
    def __apply_iteration_logic__(self):
        if self.CURRENT < self.END:
            self.CURRENT = min(self.END, self.CURRENT + self.STEP_SIZE)
            return True
        
        self.CURRENT = self.START
        return False
    
    def reset(self):
        self.CURRENT = self.START
        return super().reset()

class LayerDefinition(AGenesisTaggable):
    def __init__(self, subnet_descendants : IterativeParameterValue, switch_count : IterativeParameterValue, max_hosts_per_switch : IterativeParameterValue, repetitions : IterativeParameterValue, host_types : dict, structure_distribution : dict, layer_classification = ENetworkLayer.AGGREGATED_CONTROL, use_latex_tag = False):
        super().__init__(use_latex_tag)
        self.subnet_descendants = subnet_descendants
        self.switch_count = switch_count
        self.max_hosts_per_switch = max_hosts_per_switch
        self.host_types = host_types
        self.structure_distribution = structure_distribution
        self.layer_classification = layer_classification
        self.repetitions = repetitions

    @staticmethod
    def from_dict(value : dict, use_latex_tag = False, layer_type = ENetworkLayer.AGGREGATED_CONTROL):
        subnet_descendants = IterativeParameterValue.from_dict(value.get(SUBNET_DESCENDANTS_KEY, 1), minimum=1, use_latex_tag=use_latex_tag)
        switch_count = IterativeParameterValue.from_dict(value.get(SWITCH_COUNT_KEY, 1), minimum=-1, use_latex_tag=use_latex_tag)
        max_hosts_per_switch = IterativeParameterValue.from_dict(value.get(MAX_HOSTS_PER_SWITCH_KEY, 1), minimum=1, use_latex_tag=use_latex_tag)
        repetitions = IterativeParameterValue.from_dict(value.get(REPETITIONS_KEY, 1), minimum=1, use_latex_tag=use_latex_tag)
        host_types : dict = value.get(HOST_TYPES_KEY, dict())
        host_types = {role : int(get_single_value_by_multiple_keys(host_types, [role.name, role.name[:2], role.value], 0)) for role in EDeviceRole.configurables()}
        structure_distribution = value.get(STRUCTURE_KEY, dict())
        structure_distribution = {key : get_single_value_by_multiple_keys(structure_distribution, [key.name, key.name[0]], 0) for key in ESubnetTopologyStructure}
        if sum(structure_distribution.values()) == 0:
            structure_distribution[ESubnetTopologyStructure.LINE] = 1
        return LayerDefinition(
            subnet_descendants=subnet_descendants,
            switch_count=switch_count,
            max_hosts_per_switch=max_hosts_per_switch,
            host_types=host_types,
            structure_distribution=structure_distribution,
            layer_classification=layer_type,
            repetitions=repetitions,
            use_latex_tag=use_latex_tag)
    
    @staticmethod
    def from_str(value : str, use_latex_tag = False, layer_type = ENetworkLayer.AGGREGATED_CONTROL):
        def parse_host_types(raw_host_types : str):
            if raw_host_types == UNSET_INDICATOR:
                return {key : 0 for key in EDeviceRole.configurables()}
            
            host_types = {}
            for i, snippet in enumerate(raw_host_types.split(SMALL_SEPARATOR)):
                if match(r'(^-?\d+)', snippet):
                    host_types[EDeviceRole.configurables()[i]] = int(snippet)
                else:
                    role = EDeviceRole.from_configurables_id(snippet)
                    host_types[role] = -1

                    found = search(r'(-?\d+)', snippet)
                    if found:
                        host_types[role] = int(found.group())
            return {key : host_types.get(key, 0) for key in EDeviceRole.configurables()}
            
        def parse_structures(raw_structures : str):
            if raw_structures == UNSET_INDICATOR:
                return {key : 0 if key != ESubnetTopologyStructure.LINE else 1 for key in ESubnetTopologyStructure}
            structures = {}
            for i, snippet in enumerate(raw_structures.split(SMALL_SEPARATOR)):
                if match(r'(^-?\d+)', snippet):
                    structures[list(ESubnetTopologyStructure)[i]] = int(snippet)
                else:
                    subnet_type = ESubnetTopologyStructure.from_str(snippet[0])
                    structures[subnet_type] = 1

                    found = search(r'(^\d+)', snippet)
                    if found:
                        structures[subnet_type] = int(found.group())

            return {key : structures.get(key, 0) for key in ESubnetTopologyStructure}

        preamble, raw_host_types, raw_structures = value.split(MEDIUM_SEPARATOR)
        subnet_descendants, switch_counts, max_hosts_per_switch, repetitions = [IterativeParameterValue(1, use_latex_tag=use_latex_tag) if x == UNSET_INDICATOR else IterativeParameterValue.from_str(x, use_latex_tag=use_latex_tag) for x in preamble.split(SMALL_SEPARATOR)]

        return LayerDefinition(
            subnet_descendants=subnet_descendants,
            switch_count=switch_counts,
            max_hosts_per_switch=max_hosts_per_switch,
            host_types=parse_host_types(raw_host_types),
            structure_distribution=parse_structures(raw_structures),
            layer_classification=layer_type,
            repetitions=repetitions,
            use_latex_tag=use_latex_tag
        )

    def __get_static_str_suffix__(self):
        if self.use_latex_tag:
            host_types = SMALL_SEPARATOR.join([str(self.host_types.get(x, 0)) for x in EDeviceRole.configurables()])
            structures = SMALL_SEPARATOR.join([str(self.structure_distribution.get(x, 0)) for x in ESubnetTopologyStructure])
            return MEDIUM_SEPARATOR.join([host_types, structures])


        host_types = []
        for key, value in self.host_types.items():
            if value != 0:
                if value < 0:
                    host_types.append(key.value[0])
                else:
                    host_types.append(f'{key.value[0]}{value}')
        host_types = SMALL_SEPARATOR.join(host_types) if len(host_types) > 0 else UNSET_INDICATOR
        structures = [f'{key.name[0]}{value if value > 1 else ""}' for key, value in self.structure_distribution.items() if value > 0]
        structures = SMALL_SEPARATOR.join(structures) if len(structures) > 0 else UNSET_INDICATOR
        return MEDIUM_SEPARATOR.join([host_types, structures])
    
    def to_run_str(self):
        values = [x.to_run_str(default_value=1) for x in [self.subnet_descendants, self.switch_count, self.max_hosts_per_switch, self.repetitions]]
        if (not self.use_latex_tag) and all(x == UNSET_INDICATOR for x in values):
            preamble = UNSET_INDICATOR
        else:
            preamble = SMALL_SEPARATOR.join(values)
        return MEDIUM_SEPARATOR.join([preamble, self.__get_static_str_suffix__()])

    def to_total_str(self):
        preamble = SMALL_SEPARATOR.join(map(lambda x: x.to_total_str(default_value=1), [self.subnet_descendants, self.switch_count, self.max_hosts_per_switch, self.repetitions]))
        return MEDIUM_SEPARATOR.join([preamble, self.__get_static_str_suffix__()])

    def to_run_dict_value(self):
        return {
            SUBNET_DESCENDANTS_KEY : self.subnet_descendants.to_run_dict_value(),
            SWITCH_COUNT_KEY : self.switch_count.to_run_dict_value(),
            MAX_HOSTS_PER_SWITCH_KEY : self.max_hosts_per_switch.to_run_dict_value(),
            HOST_TYPES_KEY : {key.name : value for key, value in self.host_types.items()},
            STRUCTURE_KEY : {key.name : value for key, value in self.structure_distribution.items()},
            REPETITIONS_KEY : self.repetitions.to_run_dict_value()
        }
    
    def to_total_dict_value(self):
        return {
            SUBNET_DESCENDANTS_KEY : self.subnet_descendants.to_total_dict_value(),
            SWITCH_COUNT_KEY : self.switch_count.to_total_dict_value(),
            MAX_HOSTS_PER_SWITCH_KEY : self.max_hosts_per_switch.to_total_dict_value(),
            HOST_TYPES_KEY : {key.name : value for key, value in self.host_types.items()},
            STRUCTURE_KEY : {key.name : value for key, value in self.structure_distribution.items()},
            REPETITIONS_KEY : self.repetitions.to_total_dict_value()
        }
    
    def __apply_iteration_logic__(self):
        for iterable_value in [self.subnet_descendants, self.switch_count, self.max_hosts_per_switch, self.repetitions]:
            if iterable_value.iterate():
                return True

        return False
    
    def reset(self):
        for iterable_value in [self.subnet_descendants, self.switch_count, self.max_hosts_per_switch, self.repetitions]:
            iterable_value.reset()

        return super().reset()

class TopologyGenerationConfig(AGenesisTaggable):

    def __init__(self, iterations : int, layer_definitions : list[LayerDefinition], use_latex_tag = False):
        super().__init__(use_latex_tag)
        self.iterations = iterations
        self.layer_definitions = layer_definitions

    @staticmethod
    def from_dict(value : dict, use_latex_tag = False):
        iterations = value.get(ITERATIONS_KEY, 1)
        layer_definitions = [LayerDefinition.from_dict(layer_definition, use_latex_tag=use_latex_tag) for layer_definition in value.get(LAYER_DEFINITIONS_KEY, [])]
        if len(layer_definitions) > 0:
            layer_definitions[0].layer_classification = ENetworkLayer.CONNECTIVITY
            layer_definitions[-1].layer_classification = ENetworkLayer.PROCESS

        return TopologyGenerationConfig(iterations=iterations,
                                        layer_definitions=layer_definitions,
                                        use_latex_tag=use_latex_tag)
    
    @staticmethod
    def from_str(value : list[str], use_latex_tag = False):
        iterations = int(value[0]) if value[0] != UNSET_INDICATOR else 1
        layer_definitions = [LayerDefinition.from_str(layer_definition, use_latex_tag=use_latex_tag) for layer_definition in value[1:]] if len(value) > 1 else []
        if len(layer_definitions) > 0:
            layer_definitions[0].layer_classification = ENetworkLayer.CONNECTIVITY
            layer_definitions[-1].layer_classification = ENetworkLayer.PROCESS
        return TopologyGenerationConfig(iterations=iterations,
                                        layer_definitions=layer_definitions,
                                        use_latex_tag=use_latex_tag)

    def to_run_str(self):
        iterations = str(1)
        if not self.use_latex_tag:
            iterations = UNSET_INDICATOR
        return BIG_SEPARATOR.join([iterations] + [x.to_run_str() for x in self.layer_definitions])
    
    def to_total_str(self):
        iterations = str(self.iterations)
        if (not self.use_latex_tag) and self.iterations == 1:
            iterations = UNSET_INDICATOR
        return BIG_SEPARATOR.join([iterations] + [x.to_total_str() for x in self.layer_definitions])
    
    def to_run_dict_value(self):
        return {
            ITERATIONS_KEY : 1,
            LAYER_DEFINITIONS_KEY : [layer_definition.to_run_dict_value() for layer_definition in self.layer_definitions]
        }
    
    def to_total_dict_value(self):
        return {
            ITERATIONS_KEY : self.iterations,
            LAYER_DEFINITIONS_KEY : [layer_definition.to_total_dict_value() for layer_definition in self.layer_definitions]
        }
    
    def __apply_iteration_logic__(self):
        for layer_definition in self.layer_definitions:
            if layer_definition.iterate():
                return True
        return False
    
    def reset(self):
        for layer_definition in self.layer_definitions:
            layer_definition.reset()
        return super().reset()
    
class CommunicationGenerationConfig(AGenesisTaggable):
    DEFAULT_MAPPING = [1,0,-1]

    def __init__(self, iterations : int = DEFAULT_MAPPING[0], traffic_profile : ETrafficProfile = ETrafficProfile.from_value(DEFAULT_MAPPING[1]), connection_bound : IterativeParameterValue = None, use_latex_tag = False):
        super().__init__(use_latex_tag)
        self.iterations = iterations
        self.traffic_profile = traffic_profile
        if connection_bound:
            self.connection_bound = connection_bound
        else:
            self.connection_bound = IterativeParameterValue(CommunicationGenerationConfig.DEFAULT_MAPPING[2], use_latex_tag=use_latex_tag)

    @staticmethod
    def from_str(value : str, use_latex_tag = False):
        if value == UNSET_INDICATOR:
            iterations, traffic_profile_id, connection_bound = CommunicationGenerationConfig.DEFAULT_MAPPING.copy()
        else:
            iterations, traffic_profile_id, connection_bound = [(CommunicationGenerationConfig.DEFAULT_MAPPING[i] if x == UNSET_INDICATOR else int(x)) for (i, x) in enumerate(value.split(SMALL_SEPARATOR))]
        
        return CommunicationGenerationConfig(iterations=(1 if iterations == UNSET_INDICATOR else int(iterations)),
                                             traffic_profile=ETrafficProfile.from_value(int(traffic_profile_id)),
                                             connection_bound=IterativeParameterValue(connection_bound, use_latex_tag=use_latex_tag),
                                             use_latex_tag=use_latex_tag)

    @staticmethod
    def from_dict(value : dict, use_latex_tag = False):
        iterations = value.get(ITERATIONS_KEY, 1)
        traffic_profile = ETrafficProfile.from_str(value.get(TRAFFIC_PROFILE_KEY, ETrafficProfile.STRICT_ISOLATION.name))
        connection_bound = IterativeParameterValue.from_dict(value.get(UPPER_CONNECTION_BOUND_KEY, -1), use_latex_tag=use_latex_tag)
        return CommunicationGenerationConfig(iterations=iterations,
                                             traffic_profile=traffic_profile,
                                             connection_bound=connection_bound,
                                             use_latex_tag=use_latex_tag)
    
    def to_run_str(self):
        if (not self.use_latex_tag) and all(int(x) == CommunicationGenerationConfig.DEFAULT_MAPPING[i] for (i,x) in enumerate([1, self.traffic_profile.value, self.connection_bound.CURRENT])):
            return UNSET_INDICATOR
        
        traffic_profile = str(self.traffic_profile.value)
        if (not self.use_latex_tag) and self.traffic_profile.value == CommunicationGenerationConfig.DEFAULT_MAPPING[1]:
            traffic_profile = UNSET_INDICATOR

        connection_bound = self.connection_bound.to_run_str(default_value=CommunicationGenerationConfig.DEFAULT_MAPPING[2])
        iterations = str(self.iterations) if self.use_latex_tag else UNSET_INDICATOR
        return SMALL_SEPARATOR.join([iterations, traffic_profile, connection_bound])
    
    def to_total_str(self):
        if (not self.use_latex_tag) and all(x == CommunicationGenerationConfig.DEFAULT_MAPPING[i] for (i,x) in enumerate([self.iterations, self.traffic_profile.value, self.connection_bound.to_total_dict_value()])):
            return UNSET_INDICATOR
        
        iterations = str(self.iterations)
        if (not self.use_latex_tag) and self.iterations == CommunicationGenerationConfig.DEFAULT_MAPPING[0]:
            iterations = UNSET_INDICATOR

        traffic_profile = str(self.traffic_profile.value)
        if (not self.use_latex_tag) and self.traffic_profile.value == CommunicationGenerationConfig.DEFAULT_MAPPING[1]:
            traffic_profile = UNSET_INDICATOR
            
        connection_bound = self.connection_bound.to_total_str(CommunicationGenerationConfig.DEFAULT_MAPPING[2])
        return SMALL_SEPARATOR.join([iterations, traffic_profile, connection_bound])

    def to_run_dict_value(self):
        return {
            ITERATIONS_KEY : 1,
            TRAFFIC_PROFILE_KEY : self.traffic_profile.name,
            UPPER_CONNECTION_BOUND_KEY : self.connection_bound.to_run_dict_value()
        }
    
    def to_total_dict_value(self):
        return {
            ITERATIONS_KEY : self.iterations,
            TRAFFIC_PROFILE_KEY : self.traffic_profile.name,
            UPPER_CONNECTION_BOUND_KEY : self.connection_bound.to_total_dict_value()
        }
    
    def __apply_iteration_logic__(self):
        if self.connection_bound.iterate():
            return True
        return False
    
    def reset(self):
        self.connection_bound.reset()
        return super().reset()
    
class SecurityGenerationConfig(AGenesisTaggable):
    DEFAULT_MAPPING = [1,0,0]

    def __init__(self, iterations : int = DEFAULT_MAPPING[0], ruleset_anomaly_count = None, stateful_rule_percentage = None, use_latex_tag = False):
        super().__init__(use_latex_tag)
        self.iterations = iterations
        if ruleset_anomaly_count:
            self.ruleset_anomaly_count = ruleset_anomaly_count
        else:
            self.ruleset_anomaly_count = IterativeParameterValue(SecurityGenerationConfig.DEFAULT_MAPPING[1], use_latex_tag=use_latex_tag)

        if stateful_rule_percentage:
            self.stateful_percentage = stateful_rule_percentage
        else:
            self.stateful_percentage = IterativeParameterValue(SecurityGenerationConfig.DEFAULT_MAPPING[2], use_latex_tag=use_latex_tag)

    @staticmethod
    def from_dict(value : dict, use_latex_tag = False):
        iterations = value.get(ITERATIONS_KEY, 1)
        ruleset_anomaly_count = IterativeParameterValue.from_dict(value.get(ANOMALY_COUNT_KEY, 0), use_latex_tag=use_latex_tag)
        stateful_rule_percentage = IterativeParameterValue.from_dict(value.get(STATEFUL_PERCENTAGE_KEY, 0), minimum=0, maximum=100, use_latex_tag=use_latex_tag)
        return SecurityGenerationConfig(iterations=iterations,
                                        ruleset_anomaly_count=ruleset_anomaly_count,
                                        stateful_rule_percentage=stateful_rule_percentage,
                                        use_latex_tag=use_latex_tag)
    
    @staticmethod
    def from_str(value : str, use_latex_tag = False):
        if value == UNSET_INDICATOR:
            iterations, ruleset_anomaly_count, stateful_rule_percentage = [str(x) for x in SecurityGenerationConfig.DEFAULT_MAPPING]
        else:
            iterations, ruleset_anomaly_count, stateful_rule_percentage = [str(SecurityGenerationConfig.DEFAULT_MAPPING[i]) if x == UNSET_INDICATOR else x for (i,x) in enumerate(value.split(SMALL_SEPARATOR))]
        ruleset_anomaly_count = IterativeParameterValue.from_str(ruleset_anomaly_count, minimum=0, use_latex_tag=use_latex_tag)
        stateful_rule_percentage = IterativeParameterValue.from_str(stateful_rule_percentage, minimum=0, maximum=100, use_latex_tag=use_latex_tag)
        return SecurityGenerationConfig(iterations=int(iterations),
                                        ruleset_anomaly_count=ruleset_anomaly_count,
                                        stateful_rule_percentage=stateful_rule_percentage,
                                        use_latex_tag=use_latex_tag)

    def to_run_str(self):
        if (not self.use_latex_tag) and all((x.to_run_dict_value() == SecurityGenerationConfig.DEFAULT_MAPPING[i + 1]) for (i, x) in enumerate([self.ruleset_anomaly_count, self.stateful_percentage])):
            return UNSET_INDICATOR
        
        iterations = str(1)
        if not self.use_latex_tag:
            iterations = UNSET_INDICATOR

        ruleset_anomaly_count = self.ruleset_anomaly_count.to_run_str(SecurityGenerationConfig.DEFAULT_MAPPING[1])
        stateful_rule_percentage = self.stateful_percentage.to_run_str(SecurityGenerationConfig.DEFAULT_MAPPING[2])
        return SMALL_SEPARATOR.join([iterations, ruleset_anomaly_count, stateful_rule_percentage])
    
    def to_total_str(self):
        if (not self.use_latex_tag) and all((x == SecurityGenerationConfig.DEFAULT_MAPPING[i]) for (i, x) in enumerate([self.iterations, self.ruleset_anomaly_count.to_total_dict_value(), self.stateful_percentage.to_total_dict_value()])):
            return UNSET_INDICATOR
        
        iterations = str(self.iterations)
        if (not self.use_latex_tag) and self.iterations == SecurityGenerationConfig.DEFAULT_MAPPING[0]:
            iterations = UNSET_INDICATOR
        ruleset_anomaly_count = self.ruleset_anomaly_count.to_total_str(SecurityGenerationConfig.DEFAULT_MAPPING[1])
        stateful_rule_percentage = self.stateful_percentage.to_total_str(SecurityGenerationConfig.DEFAULT_MAPPING[2])
        return SMALL_SEPARATOR.join([iterations, ruleset_anomaly_count, stateful_rule_percentage])

    def to_run_dict_value(self):
        return {
            ITERATIONS_KEY : 1,
            ANOMALY_COUNT_KEY : self.ruleset_anomaly_count.to_run_dict_value(),
            STATEFUL_PERCENTAGE_KEY : self.stateful_percentage.to_run_dict_value()
        }
    
    def to_total_dict_value(self):
        return {
            ITERATIONS_KEY : self.iterations,
            ANOMALY_COUNT_KEY : self.ruleset_anomaly_count.to_total_dict_value(),
            STATEFUL_PERCENTAGE_KEY : self.stateful_percentage.to_total_dict_value()
        }
    
    def __apply_iteration_logic__(self):
        for iterable_value in [self.ruleset_anomaly_count, self.stateful_percentage]:
            if iterable_value.iterate():
                return True

        return False
    
    def reset(self):
        for iterable_value in [self.ruleset_anomaly_count, self.stateful_percentage]:
            iterable_value.reset()
        return super().reset()
        
class SeedConfig(AGenesisTaggable):
    def __init__(self, topology_seed : IterativeParameterValue, communication_seed : IterativeParameterValue, security_seed : IterativeParameterValue, use_latex_tag = False):
        super().__init__(use_latex_tag)
        self.topology_seed = topology_seed
        self.communication_seed = communication_seed
        self.security_seed = security_seed

    @staticmethod
    def from_str(value : str, iterations = None, use_latex_tag = False):
        topology_seed, communication_seed, security_seed = [IterativeParameterValue(start=int(x), end=(int(x) + iterations[i] - 1) if iterations else None, use_latex_tag=use_latex_tag) for i, x in enumerate(value.split(SMALL_SEPARATOR))]
        return SeedConfig(topology_seed=topology_seed,
                          communication_seed=communication_seed,
                          security_seed=security_seed)
    
    @staticmethod
    def from_dict(value : dict, iterations = None, use_latex_tag = False):
        raw_seeds = [value.get(key, randint(1,1000)) for key in [TOPOLOGY_KEY, COMMUNICATION_KEY, SECURITY_KEY]]
        topology_seed, communication_seed, security_seed = [IterativeParameterValue(start=seed, end=(seed + iterations[i] - 1) if iterations else None, use_latex_tag=use_latex_tag) for i, seed in enumerate(raw_seeds)]

        return SeedConfig(topology_seed=topology_seed,
                          communication_seed=communication_seed,
                          security_seed=security_seed)
    
    @staticmethod
    def random(iterations = None, use_latex_tag = False):
        raw_seeds = [randint(1,1000) for _ in range(3)]
        topology_seed, communication_seed, security_seed = [IterativeParameterValue(start=seed, end=(seed + iterations[i] - 1) if iterations else None, use_latex_tag=use_latex_tag) for i, seed in enumerate(raw_seeds)]
    
        return SeedConfig(topology_seed=topology_seed,
                          communication_seed=communication_seed,
                          security_seed=security_seed)

    def __apply_iteration_logic__(self):
        for iterative_value in [self.topology_seed, self.communication_seed, self.security_seed]:
            if iterative_value.iterate():
                return True
        return False
    
    def reset(self):
        for iterative_value in [self.topology_seed, self.communication_seed, self.security_seed]:
            iterative_value.reset()
        return super().reset()

    def to_run_str(self):
        return SMALL_SEPARATOR.join([str(x.CURRENT) for x in [self.topology_seed, self.communication_seed, self.security_seed]])
    
    def to_total_str(self):
        return SMALL_SEPARATOR.join([str(x.START) for x in [self.topology_seed, self.communication_seed, self.security_seed]])
    
    def to_run_dict_value(self):
        return {key : value.CURRENT for key, value in [(TOPOLOGY_KEY, self.topology_seed), (COMMUNICATION_KEY, self.communication_seed), (SECURITY_KEY, self.security_seed)]}
    
    def to_total_dict_value(self):
        return {key : value.START for key, value in [(TOPOLOGY_KEY, self.topology_seed), (COMMUNICATION_KEY, self.communication_seed), (SECURITY_KEY, self.security_seed)]}
    
class GenerationConfig(AGenesisTaggable):
    def __init__(self, seed_config : SeedConfig, topology_config : TopologyGenerationConfig, communication_config : CommunicationGenerationConfig, security_config : SecurityGenerationConfig, use_latex_tag = False):
        super().__init__(use_latex_tag)
        self.seed_config = seed_config
        self.topology_config = topology_config
        self.communication_config = communication_config
        self.security_config = security_config

    @staticmethod
    def from_str(genesis_tag : str, use_latex_tag = False):
        versioning, information, _ = genesis_tag.split(TAG_ENCAPSULATOR)
        version_number = versioning.split(MEDIUM_SEPARATOR)[-1][1:]
        if version(GENESIS_PACKAGE_NAME) != version_number:
            raise Exception(f'InvalidGenesisTag: Incompatible GeNESIS version detected: {version(GENESIS_PACKAGE_NAME)} != {version_number}. You can checkout the GeNESIS version matching your tag with "git checkout tags/{version_number}" *if it exists*.')
        sequences = information.split(BIG_SEPARATOR)
        communication_config = CommunicationGenerationConfig.from_str(sequences[1], use_latex_tag=use_latex_tag)
        security_config = SecurityGenerationConfig.from_str(sequences[2], use_latex_tag=use_latex_tag)
        topology_config = TopologyGenerationConfig.from_str(sequences[3:], use_latex_tag=use_latex_tag)
        seed_config = SeedConfig.from_str(sequences[0], [topology_config.iterations, communication_config.iterations, security_config.iterations], use_latex_tag=use_latex_tag)
        return GenerationConfig(seed_config=seed_config,
                                topology_config=topology_config,
                                communication_config=communication_config,
                                security_config=security_config,
                                use_latex_tag=use_latex_tag)

    @staticmethod
    def from_dict(value : dict, use_latex_tag = False):
        topology_config = TopologyGenerationConfig.from_dict(value.get(TOPOLOGY_KEY, dict()), use_latex_tag=use_latex_tag)

        raw_communication_config = value.get(COMMUNICATION_KEY, dict())
        if raw_communication_config:
            communication_config = CommunicationGenerationConfig.from_dict(raw_communication_config, use_latex_tag=use_latex_tag)
        else:
            communication_config = CommunicationGenerationConfig(use_latex_tag=use_latex_tag)
        
        raw_security_config = value.get(SECURITY_KEY, dict())
        if raw_security_config:
            security_config = SecurityGenerationConfig.from_dict(raw_security_config, use_latex_tag=use_latex_tag)
        else:
            security_config = SecurityGenerationConfig(use_latex_tag=use_latex_tag)
            
        return GenerationConfig(seed_config=SeedConfig.random([topology_config.iterations, communication_config.iterations, security_config.iterations], use_latex_tag=use_latex_tag),
                                topology_config=topology_config,
                                communication_config=communication_config,
                                security_config=security_config,
                                use_latex_tag=use_latex_tag)

    @staticmethod
    def from_file(dir : str, use_latex_tag = False):
        with open(dir, 'r') as file:
            json_config : dict = load(file)
            return GenerationConfig.from_dict(json_config, use_latex_tag)

    
    def to_run_str(self):
        content = BIG_SEPARATOR.join(map(lambda x: x.to_run_str(), [self.seed_config, self.communication_config, self.security_config, self.topology_config]))
        if self.use_latex_tag:
            return f'genesis:v{version(GENESIS_PACKAGE_NAME)}\\{TAG_ENCAPSULATOR}{content}\\{TAG_ENCAPSULATOR}'
        return f'genesis:v{version(GENESIS_PACKAGE_NAME)}{TAG_ENCAPSULATOR}{content}{TAG_ENCAPSULATOR}'
    
    def to_total_str(self):
        content = BIG_SEPARATOR.join(map(lambda x: x.to_total_str(), [self.seed_config, self.communication_config, self.security_config, self.topology_config]))
        if self.use_latex_tag:
            return f'genesis:v{version(GENESIS_PACKAGE_NAME)}\\{TAG_ENCAPSULATOR}{content}\\{TAG_ENCAPSULATOR}'
        return f'genesis:v{version(GENESIS_PACKAGE_NAME)}{TAG_ENCAPSULATOR}{content}{TAG_ENCAPSULATOR}'

    def to_run_dict_value(self):
        return {key : value.to_run_dict_value() for key, value in [(TOPOLOGY_KEY, self.topology_config), (COMMUNICATION_KEY, self.communication_config), (SECURITY_KEY, self.security_config)]}
    
    def to_total_dict_value(self):
        return {key : value.to_total_dict_value() for (key, value) in [(TOPOLOGY_KEY, self.topology_config), (COMMUNICATION_KEY, self.communication_config), (SECURITY_KEY, self.security_config)]}

    def to_run_file(self, location):
        with open(location, 'x') as file:
            dump(self.to_run_dict_value(), file, indent=4)
        print('Configuration saved at: ', location)

    def to_total_file(self, location):
        with open(location, 'x') as file:
            dump(self.to_total_dict_value(), file, indent=4)
        print('Configuration saved at: ', location)

    def __apply_iteration_logic__(self):
        for subconfig in [self.topology_config, self.communication_config, self.security_config]:
            if subconfig.iterate():
                return True
        return False
    
    def reset(self):
        for subconfig in [self.topology_config, self.communication_config, self.security_config]:
            subconfig.reset()
        return super().reset()


from copy import copy
from math import ceil

from hses_genesis.generation.communication import CommunicationGenerator
from hses_genesis.generation.dynamic_configuration import ConfigurationGenerator
from hses_genesis.generation.network_configuration import NetworkConfigurationGenerator
from hses_genesis.generation.topology import TopologyGenerator, calculate_subnets, get_graph_information
from hses_genesis.parsing.configuration import GenerationConfig
from hses_genesis.setup.output_location import setup_base_location, setup_run_location
from hses_genesis.utils.constants import GRAPH_FOLDER, RULESET_FOLDER, TERMINAL_WIDTH
from hses_genesis.utils.enum_objects import EDeviceRole
from hses_genesis.save import topology as topology
from hses_genesis.save import packets as packets
from hses_genesis.save import rulesets as rulesets
from argparse import ArgumentParser
from os.path import abspath, dirname, join, basename, exists
from networkx import shortest_simple_paths
from hses_genesis.utils.functions import load_resource, print_information

def user_choice(prompt, choices, default = None):
    while True:
        value = input(prompt + ' (' + ', '.join(choices + ([f'default {default}'] if default else [])) + '): ')
        if not value and default:
            return default
        if value in choices:
            return value
        
        print('No valid option chosen. Please try again.')

def perform_generation_cycle(base_location : str, config : GenerationConfig, export_yang_files = False, export_iptables_files = False, export_graph_images = False, export_omnet_files = False, export_ns3_files = False, export_zimpl_parsables = False):
    print(f'Start Generation Step "topology"')
    topology_generator = TopologyGenerator(config.seed_config.topology_seed.CURRENT)
    layer_definitions = [clone for layer_definition in config.topology_config.layer_definitions for clone in [copy(layer_definition) for _ in range(layer_definition.repetitions.CURRENT)]]
    available_subnets = calculate_subnets(layer_definitions)
    G = topology_generator.generate_network(layer_definitions, available_subnets)
    print_information('GRAPH INFORMATION', get_graph_information(G))

    print(f'Start Generation Step "communication"')
    communication_generator = CommunicationGenerator(config.seed_config.communication_seed.CURRENT)
    _, sampled_connections = communication_generator.get_connections(G, config.communication_config.traffic_profile, config.communication_config.connection_bound.CURRENT)
    
    ruleset_connections = []
    generated_packets = {}

    for (source_id, destination_id, service) in sampled_connections:
        protocol = communication_generator.random.choice(service.value.protocols)
        port = communication_generator.random.choice(service.value.ports)
        ruleset_connections.append(((source_id, destination_id, protocol, port, port), list(shortest_simple_paths(G, source_id, destination_id))))

        if source_id not in generated_packets:
            generated_packets[source_id] = []

        src_data, dst_data = [G.nodes[node] for node in [source_id, destination_id]]

        src_packet = communication_generator.generate_packet(service=service,
                                                src_node=source_id,
                                                dst_node=destination_id,
                                                src_data=src_data,
                                                dst_data=dst_data,
                                                protocol=protocol,
                                                port=port,
                                                is_high_sender=src_data['role'] in EDeviceRole.high_senders())
        
        generated_packets[source_id].append(src_packet)

        if destination_id not in generated_packets:
            generated_packets[destination_id] = []
            
        dst_packet = communication_generator.generate_packet(service=service,
                                                src_node=destination_id,
                                                dst_node=source_id,
                                                src_data=dst_data,
                                                dst_data=src_data,
                                                protocol=protocol,
                                                port=port,
                                                is_high_sender=dst_data['role'] in EDeviceRole.high_senders())
        
        generated_packets[destination_id].append(dst_packet)

    run_location = setup_run_location(config=config,
                                      base_location=base_location,
                                      run_label='-'.join(map(lambda x: str(x), [config.seed_config.topology_seed.CURRENT, config.seed_config.communication_seed.CURRENT, config.seed_config.security_seed.CURRENT])),
                                      export_iptables_files = export_iptables_files,
                                      export_omnet_files = export_omnet_files,
                                      export_ns3_files = export_ns3_files,
                                      export_zimpl_parsables = export_zimpl_parsables)
    
    packets.to_csv(run_location, generated_packets)
    
    print(f'Start Generation Step "security"')
    network_configuration_generator = NetworkConfigurationGenerator(config.seed_config.security_seed.CURRENT)

    routers = [router for router, data in G.nodes(data=True) if data['role'] == EDeviceRole.ROUTER]

    router_connections = {}
    for (source, target, p, sport, dport), paths in ruleset_connections:
        affected_routers = [router for router in routers if any(router in path for path in paths)]
        for router in affected_routers:
            if router not in router_connections.keys():
                router_connections[router] = []
            router_connections[router].append((G.nodes[source]['ip'], G.nodes[target]['ip'], p, sport, dport))

    router_ruleset_map = {}
    for router in router_connections.keys():
        raw_ruleset = network_configuration_generator.generate_ruleset(set(router_connections[router]), config.security_config.ruleset_anomaly_count.CURRENT, config.security_config.stateful_percentage.CURRENT)
        G.nodes[router]['ruleset'] = [NetworkConfigurationGenerator.rule_to_str(rule) for rule in raw_ruleset]
        if export_iptables_files:
            rulesets.to_save_file(join(run_location, RULESET_FOLDER), router, G.nodes[router]['ruleset'])
        router_ruleset_map[router] = [NetworkConfigurationGenerator.to_numerical_representation(rule) for rule in raw_ruleset]
    
    print_information('GENERAL INFORMATION', {
        'seeds used for this iteration' : '',
        '- topology' : str(config.seed_config.topology_seed.CURRENT),
        '- communication' : str(config.seed_config.communication_seed.CURRENT),
        '- security' : str(config.seed_config.security_seed.CURRENT),
        'output data location for this iteration' : run_location,
        'genesis tag for this iteration' : config.to_run_str()
    })

    topology.to_graphml(G, join(run_location, GRAPH_FOLDER))
    if export_yang_files:
        topology.to_ietf(G, join(run_location, GRAPH_FOLDER), generated_packets)
    if export_graph_images:
        topology.to_image(G, join(run_location, GRAPH_FOLDER), seed=config.seed_config.topology_seed.CURRENT)
    if export_omnet_files:
        topology.to_omnet_ned(G, run_location)
        packets.to_omnet_ini(run_location, generated_packets)
    if export_ns3_files:
        topology.to_ns3_cc(G, run_location, generated_packets)
    if export_zimpl_parsables:
        topology.to_zimpl_parsable(G, run_location)

parser = ArgumentParser()
parser.add_argument('-j', '--json', help='pass the name or absolute path of the configuration file to use.', default=None)
parser.add_argument('-g', '--genesis_tag', help='pass the GeNESIS-TAG of a previous run.', default=None)
parser.add_argument('-n', '--new_configuration', help='start the Interactive GeNESIS Configuration Generator to create a new configuration.', action='store_true')
parser.add_argument('-o', '--output_location', help='set the output location for generated files.', default=join(dirname(abspath(__file__)), 'output'))
parser.add_argument('-img', '--export_graph_images', help='export a .png and a .jpg of the network topology.', action='store_true')
parser.add_argument('-zpl', '--export_zimpl_parsables', help='export the topology and rules as zimpl parsable txt files.', action='store_true')
parser.add_argument('-omnet', '--export_omnet_files', help='export the topology and packet configuration files for omnet++.', action='store_true')
parser.add_argument('-ns3', '--export_ns3_files', help='export the topology and packet configuration files for ns3.', action='store_true')
parser.add_argument('-yang', '--export_yang_files', help='export the all outputs in a single json file.', action='store_true')
parser.add_argument('-ipt', '--export_iptables_files', help='export the scurity configurations as iptables save files.', action='store_true')
parser.add_argument('-latag', '--use_latex_tag', help='get the genesis tag in latex parsable form.', action='store_true')

args = parser.parse_args()

print('-' * ceil((TERMINAL_WIDTH - 12) / 2), 'Welcome to', '-' * ceil((TERMINAL_WIDTH - 12) / 2))
print('  ___     _  _ ___ ___ ___ ___     _   ___ ')
print(' / __|___| \| | __/ __|_ _/ __|_ _/ | |_  )')
print('| (_ / -_)  ` | _|\__ \| |\__ \ V / |_ / /')
print(' \___\___|_|\_|___|___/___|___/\_/|_(_)___|')
print()
print('-' * TERMINAL_WIDTH)

configuration_choices = [t for (t, v) in [('g', args.genesis_tag), ('j', args.json), ('n', args.new_configuration)] if v]
if len (configuration_choices) > 1:
    choice = user_choice('Multiple configuration origins were provided. Please choose the one you want to use', configuration_choices)
elif len(configuration_choices) == 1:
    choice = configuration_choices[0]
else:
    choice = 'r'

config : dict = {}

if choice == 'n':
    config = GenerationConfig.from_dict(ConfigurationGenerator().edit_config())
    config_name = 'custom_config'
elif choice == 'j':
    config_file = args.json
    if not str(config_file).endswith('.json'):
        config_file += '.json'
    if not exists(args.json):
        config_file = load_resource('configurations', config_file)
    config_name = basename(str(args.json)).removesuffix('.json')
    config : GenerationConfig = GenerationConfig.from_file(config_file, args.use_latex_tag)
elif choice == 'g':
    config_name = 'tag_rerun_config'
    config : GenerationConfig = GenerationConfig.from_str(args.genesis_tag, args.use_latex_tag)
else:
    config_name = 'example_config'
    config_file = load_resource('configurations', f'{config_name}.json')
    config : GenerationConfig = GenerationConfig.from_file(config_file, args.use_latex_tag)

print('-' * TERMINAL_WIDTH)
print(f'GeNESIS started with GeNESIS-TAG: {config.to_total_str()}')
print('-' * TERMINAL_WIDTH)
base_location = setup_base_location(config, args.output_location, config_name)

while True:
    config.reset()
    while True:
        perform_generation_cycle(base_location,
                                 config,
                                 export_yang_files=args.export_yang_files,
                                 export_iptables_files=args.export_iptables_files,
                                 export_graph_images=args.export_graph_images,
                                 export_ns3_files=args.export_ns3_files,
                                 export_omnet_files=args.export_omnet_files,
                                 export_zimpl_parsables=args.export_zimpl_parsables)
        if not config.iterate():
            break
    if not config.seed_config.iterate():
        break


from hses_genesis.parsing.ruleset import parse_tables

def from_file(file_dir, debug = False):
    with open(file_dir, 'r') as file:
        return parse_tables(list(map(lambda x: x.strip(), file.readlines())), False, debug)


from csv import DictWriter
from os.path import join
from hses_genesis.utils.constants import OMNET_FOLDER, PACKET_FOLDER, PACKET_HEADERS, ZIMPL_FOLDER
from hses_genesis.utils.enum_objects import EParameterKey
from hses_genesis.utils.functions import load_resource

def to_csv(dst_location, packets):
    with open(join(dst_location, PACKET_FOLDER, 'packets.csv'), 'w') as file:
        writer = DictWriter(file, PACKET_HEADERS)
        writer.writeheader()
        [writer.writerow(packet) for packet_list in packets.values() for packet in packet_list]

def to_zimpl_parsable(location, packets):
    with open(join(location, ZIMPL_FOLDER, f'zimpl_packets.txt'), 'w') as file:
        for i, parameter_key in enumerate(list(EParameterKey.__members__.values())):
            file.write(f'# <x,{i}>{parameter_key.name.lower()}\n')
        for i, packet in enumerate([packet for packet_list in packets.values() for packet in packet_list]):
            for j, key in enumerate([f'{EParameterKey.SRC.name.lower()}_ip', f'{EParameterKey.DST.name.lower()}_ip', f'{EParameterKey.PROTOCOL.name.lower()}_code', EParameterKey.SRC_PORT.value, EParameterKey.DST_PORT.value]):
                file.write(f'{i},{j},{int(packet[key])}\n')

def to_omnet_ini(location, app_map):
    packet_map = {}
    for packets in app_map.values():
        for packet in packets:
            src, dst = packet['s'], packet['d']
            if src not in packet_map.keys():
                packet_map[src] = []

            if dst not in packet_map.keys():
                packet_map[dst] = []

            if packet['p'] == 'udp':
                packet_map[src].append((packet, 'UdpBasicApp'))
                if not any(packet['d'] == p['d'] and packet['-dport'] == p['-dport'] and a == 'UdpSink' for (p, a) in packet_map[dst]):
                    packet_map[dst].append((packet, 'UdpSink'))
            else:
                packet_map[src].append((packet, 'TcpClientApp'))
                if not any(packet['d'] == p['d'] and packet['-dport'] == p['-dport'] and a == 'TcpSinkApp' for (p, a) in packet_map[dst]):
                    packet_map[dst].append((packet, 'TcpSinkApp'))

    with open(load_resource('templates', 'omnetpp.ini'), 'r') as template, open(join(location, OMNET_FOLDER, f'omnetpp.ini'), 'w') as file:
        file.writelines(template.readlines())
        file.write('\n')

        for src, apps in packet_map.items():
            if len(apps) == 0:
                continue
            file.write(f'**.{src}.numApps = {len(apps)}\n')
            for i, (packet, app_type) in enumerate(apps):
                file.write(f'**.{src}.app[{i}].typename = "{app_type}"\n')
                mappings = []

                if 'udp' in app_type.lower():

                    if app_type == 'UdpBasicApp':
                        file.write(f'**.{src}.app[{i}].stopTime = 25s\n')
                        mappings.extend([
                            ('localPort', '-sport', lambda _: -1),
                            ('sendInterval', 'packets_per_second', lambda x: f'{round(1 / x, 3)}s'),
                            ('messageLength', 'packet_size', lambda x: f'{x}B'),
                            ('destAddresses', 'd', lambda x: f'"{x}"'),
                            ('destPort', '-dport', lambda x: x),
                        ])
                    else:
                        mappings.append(
                            ('localPort', '-sport', lambda x: x)
                        )
                else:
                    if app_type == 'TcpClientApp':
                        file.write(f'**.{src}.app[{i}].source.packetData = intuniform(0,1)\n')
                        mappings.extend([
                            ('io.localPort', '-sport', lambda _: -1),
                            ('io.connectAddress', 'd', lambda x: f'"{x}"'),
                            ('io.connectPort', '-dport', lambda x: x),
                            ('source.productionInterval', 'packets_per_second', lambda x: f'{round(1 / x, 3)}s'),
                            ('source.packetLength', 'packet_size', lambda x: f'{x}B'),
                        ])
                    else:
                        mappings.append(('localPort', '-dport', lambda x: x))

                for param, packet_key, mapping_func in mappings:
                    file.write(f'**.{src}.app[{i}].{param} = {mapping_func(packet[packet_key])}\n')

                file.write('\n')


from networkx import Graph, read_graphml
from hses_genesis.parsing.ruleset import parse_tables
from hses_genesis.utils.enum_objects import EPacketDecision, EService, EDeviceRole

def from_file(file_dir, debug = False):
    G : Graph = read_graphml(file_dir)
    for n in G.nodes():
        node_info = G.nodes[n]
        if 'role' in node_info.keys():
            G.nodes[n]['role'] = EDeviceRole.from_name(G.nodes[n]['role'])
        if 'ruleset' in node_info.keys():
            tables = parse_tables(node_info['ruleset'].split('\n'), True, debug)
            G.nodes[n]['ruleset'] = [(chain, conditions, action) for (_, _, ruleset) in tables for (chain, conditions, action) in ruleset]
        if 'services' in node_info.keys():
            G.nodes[n]['services'] = list(map(lambda x: EService.from_str(x), node_info['services'].split(','))) if ',' in G.nodes[n]['services'] else []
        if 'default_action' in node_info.keys():
            G.nodes[n]['default_action'] = EPacketDecision.from_str(node_info['default_action'])
    return G


from csv import DictReader
from hses_genesis.utils.constants import PACKET_HEADERS


def from_csv(file_path):
    with open(file_path, 'r') as file:
        return list(DictReader(file, PACKET_HEADERS))[1:]


from ipaddress import ip_address, ip_network
from json import dumps
from warnings import warn
from matplotlib.pyplot import close, savefig
from networkx import Graph, draw, write_graphml, spring_layout
from os.path import join

from hses_genesis.utils.constants import NS3_FOLDER, OMNET_FOLDER, ZIMPL_FOLDER
from hses_genesis.utils.enum_objects import EDeviceRole, EParameterKey
from hses_genesis.utils.functions import load_resource

def prepare_packets(G : Graph, packets = None):
    if not packets:
        return []
    
    output = []
    for packet in packets:
        output.append(
            {
                EParameterKey.SRC.value : G.nodes(packet[EParameterKey.SRC.value])['ip'],
                EParameterKey.DST.value : G.nodes(packet[EParameterKey.DST.value])['ip'],
                EParameterKey.PROTOCOL.value : packet[EParameterKey.PROTOCOL.value],
                EParameterKey.SRC_PORT.value : packet[EParameterKey.SRC_PORT.value],
                EParameterKey.DST_PORT.value : packet[EParameterKey.DST_PORT.value],
                'packet_size' : packet['packet_size'],
                'packets_per_second' : packet['packets_per_second']
            }
        )
    return output
    
def prepare_topology(G):
    g_copy = G.copy()
    for n, data in g_copy.nodes(data = True):
        if 'services' in data.keys():
            g_copy.nodes[n]['services'] = ','.join([service.name for service in data['services']])

        if data['role'] == EDeviceRole.ROUTER:
            g_copy.nodes[n]['subnet'] = ','.join(g_copy.nodes[n]['subnet'])
            g_copy.nodes[n]['ip'] = ','.join(g_copy.nodes[n]['ip'])
        
        for key, joiner in [['ruleset', '\n'], ['default_action', None], ['role', None]]:
            if key in g_copy.nodes[n].keys():
                if joiner != None:
                    g_copy.nodes[n][key] = joiner.join([str(v) for v in g_copy.nodes[n][key]])
                else:
                    g_copy.nodes[n][key] = g_copy.nodes[n][key].name
    return g_copy

def to_ietf(G : Graph, location, generated_packets = None):
    ietf_nodes = []
    for dev in G.nodes:
        if EDeviceRole.from_device_id(dev) == EDeviceRole.PORT:
            continue

        ietf_node = {
            'node_id' : dev.lower(),
            'services' : [service.to_dict() for service, _ in G.nodes[dev]['services']]
        }
        for key in ['role', 'ip', 'subnet', 'branch', 'layer']:
            if key == 'services':
                continue
            ietf_node[key] = G.nodes[dev][key]

        for enum_key in ['default_action', 'role']:
            ietf_node[enum_key] = G.nodes[dev][enum_key].name if enum_key in G.nodes[dev] else '' 
        
        ietf_node['ruleset'] = G.nodes[dev]['ruleset'] if 'ruleset' in G.nodes[dev] else []
        ietf_node['packets'] = [] if generated_packets == None or dev not in generated_packets else prepare_packets(generated_packets[dev])
        if EDeviceRole.from_device_id(dev) not in [EDeviceRole.ROUTER, EDeviceRole.SWITCH]:
            ietf_node['termination_point'] = [f'{dev.lower()}#0']
        else:
            ietf_node['termination_point'] = [f'{dev.lower()}#{i}' for i in range(len(list(G.neighbors(dev))))]

        ietf_nodes.append(ietf_node)

    ietf_links = []
    for s, d in G.edges:
        ietf_link = {
            'source': {
                'source-node' : s.split('#')[0].lower(),
                'source-tp' : s.lower() if EDeviceRole.from_device_id(s) == EDeviceRole.PORT else f'{s.lower()}#0'
            },
            'destination' : {
                'destination-node' : d.split('#')[0].lower(),
                'source-tp' : d.lower() if EDeviceRole.from_device_id(d) == EDeviceRole.PORT else f'{d.lower()}#0'
            }
        }
        ietf_links.append(ietf_link)

    ietf_topology = {
        'ietf-network:networks' : {
        'network': [
                {
                    'network-types': {},
                    'network-id': 'ietf-topology',
                    'node': ietf_nodes,
                    'ietf-network-topology:link': ietf_links
                }
            ]
        }
    }
        
    with open(join(location, 'ietf-topology.yang.json'), 'w') as file:
        file.write(dumps(ietf_topology, indent=4))

def to_zimpl_parsable(G, location):
    zimpl_location = join(location, ZIMPL_FOLDER)
    with open(join(zimpl_location, 'vertices.txt'), 'w') as file:
        file.write('# <id,ip,is_container>\n')
        for node, data in G.nodes(data = True):
            if data['role'] == EDeviceRole.ROUTER:
                for ip in data["ip"]:
                    file.write(f'{node},{int(ip_address(ip))},{int(False)}\n')
            else:
                file.write(f'{node},{int(ip_address(data["ip"]))},{int(data["role"] == EDeviceRole.SWITCH)}\n')
    
    with open(join(zimpl_location, 'edges.txt'), 'w') as edges_file, open(join(zimpl_location, 'ip_edges.txt'), 'w') as ip_edges_file:
        edges_file.write('# <src_id,dst_id>weight\n')
        ip_edges_file.write('# <src_ip,dst_ip>weight\n')
        for (src, dst) in G.edges:
            edges_file.write(f'{src},{dst},1\n')
            src_ips, dst_ips = G.nodes[src]["ip"], G.nodes[dst]["ip"]
            src_ips = [src_ips] if not isinstance(src_ips, list) else src_ips
            dst_ips = [dst_ips] if not isinstance(dst_ips, list) else dst_ips
            for src_ip in src_ips:
                for dst_ip in dst_ips:
                    ip_edges_file.write(f'{int(ip_address(src_ip))},{int(ip_address(dst_ip))},1\n')

def to_omnet_ned(G : Graph, location):
    submodules = ['\tconfigurator: Ipv4NetworkConfigurator { }\n', '\tvisualizer: IntegratedCanvasVisualizer { }\n']
    interfaces = []
    for node, data in G.nodes(data=True):
        extension = '{ }'
        if data['role'] == EDeviceRole.ROUTER:
            role = 'Router'
        elif data['role'] == EDeviceRole.SWITCH:
            role = 'EthernetSwitch'
        else:
            role = 'StandardHost'
            extension = '{' + f'@display("i=device/pc");' + '}'

        submodules.append(f'\t\t{node}: {role} {extension}\n')

        ips = [(data['ip'], data['subnet'])] if isinstance(data['ip'], str) else [(ip, data['subnet'][i]) for i, ip in enumerate(data['ip'])]

        for ip, subnet in ips:
            interface = {
                'hosts' : node,
                'address' : ip,
                'netmask' : str(ip_network(subnet, False).netmask)
            }

            if data['role'] == EDeviceRole.ROUTER:
                target_hosts = [n for n, d in G.nodes(data=True) if n != node and d['subnet'] == subnet and d['role'] in EDeviceRole.configurables()]
                if len(target_hosts) == 0:
                    warn(f'Invalid Configuration Exception: Cannot determine interface target for {node} in interface {subnet} ({[n for n, d in G.nodes(data=True) if d["subnet"] == subnet]}). You probably defined a layer without any end devices.')
                    continue

                interface['towards'] = target_hosts[0]

            interfaces.append(interface)
            

    connections = []
    for (src, dst) in G.edges:
        connections.append(f'\t\t{src}.ethg++ <--> ethline <--> {dst}.ethg++;\n')

    lines = []
    with open(load_resource('templates', 'template.ned'), 'r') as f:
        for line in f.readlines():
            lines.append(line)
            if 'submodules:' in line:
                lines.extend(submodules)
            elif 'connections:' in line:
                lines.extend(connections)

    with open(join(location, OMNET_FOLDER, 'graph.ned'), 'w') as f:
        f.writelines(lines)

    with open(join(location, OMNET_FOLDER, 'interfaces.xml'), 'w') as f:
        f.write('<config>\n')
        for interface in interfaces:
            values = ['\t<interface']
            for key, value in interface.items():
                values.append(f'{key}="{value}"')
            f.write(' '.join(values) + '/>\n')
        f.write('</config>')

def to_graphml(G, location, file_name = 'graph.graphml'):
    write_graphml(prepare_topology(G), join(location, file_name))

def to_ns3_cc(G : Graph, location, generated_packets):

    def create_nodes(G : Graph):
        return [f'Ptr<Node> {node} = CreateObject<Node>();' for node in G.nodes]
    
    def add_names(G : Graph):
        return [f'Names::Add("{node}", {node});' for node in G.nodes]
    
    def create_links(G : Graph):
        lines = []
        for src, dst in G.edges():
            link = f'link_{src}_{dst}'
            lines.append(f'NetDeviceContainer {link} = csma.Install(NodeContainer({src}, {dst}));')
        return lines

    def create_bridges(G : Graph):
        lines = []
        links = {}

        for src, dst in G.edges():
            if G.nodes[src]['role'] != EDeviceRole.SWITCH:
                continue
            if src not in links.keys():
                links[src] = []
            links[src].append(f'link_{src}_{dst}')

        for switch, links in links.items():
            lines.append('')
            lines.append(f'NetDeviceContainer {switch}nd;')
            for link in links:
                lines.append(f'{switch}nd.Add({link}.Get(0));')
        return lines

    def install_bridges(G : Graph):
        return [f'bridge.Install({node}, {node}nd);' for node, data in G.nodes(data=True) if data['role'] == EDeviceRole.SWITCH]

    def install_end_device_ip_stack(G : Graph):
        end_points = ', '.join([node for node, data in G.nodes(data = True) if data['role'] in EDeviceRole.configurables()])
        return [f'NodeContainer endpointNodes({end_points});', 'ns3IpStack.Install(endpointNodes);']
    
    def install_router_ip_stack(G : Graph):
        routers = ', '.join([node for node, data in G.nodes(data = True) if data['role'] == EDeviceRole.ROUTER])
        return [f'NodeContainer routerNodes({routers});', 'ns3IpStack.Install(routerNodes);']
    
    def assign_router_ips(G : Graph):
        lines = []
        introduced_routers = []
        net_devices = []
        for edge in G.edges:
            routers = [(router, G.nodes[router]) for router in edge if G.nodes[router]['role'] == EDeviceRole.ROUTER]
            if len(routers) != 1:
                continue

            router, data = routers[0]
            lines.append('')

            router_occurence = int(router in introduced_routers)
            net_device_name = f'{router}_{router_occurence}'
            ipv4_name = f'ipv4proto{router}'
            interface_name = f'{router}_{router_occurence}interface'
            net_devices.append((interface_name, ipv4_name))
            
            lines.append(f'Ptr<NetDevice> {net_device_name}nd = link_{edge[0]}_{edge[1]}.Get({list(edge).index(router)});')
            
            if router not in introduced_routers:
                introduced_routers.append(router)
                lines.append(f'Ptr<Ipv4> {ipv4_name} = {router}->GetObject<Ipv4>();')

            lines.append(f'uint32_t {interface_name} = {ipv4_name}->GetInterfaceForDevice({net_device_name}nd);')
            lines.append(f'if ({interface_name} == -1)' + '{' f'{interface_name} = {ipv4_name}->AddInterface({net_device_name}nd);' + '}')
            
            subnet = G.nodes[[other for other in edge if other != router][0]]['subnet']
            ip = data['ip'][list(data['subnet']).index(subnet)]

            ipv4Mask = str(subnet).split('/')[-1]
            ip4Addr_name = f'ipv4Addr{edge[0]}{edge[1]}'
            lines.append(f'Ipv4InterfaceAddress {ip4Addr_name} = Ipv4InterfaceAddress(Ipv4Address("{ip}"), Ipv4Mask("/{ipv4Mask}"));')
            lines.append(f'{ipv4_name}->AddAddress({interface_name}, {ip4Addr_name});')

        lines.append('')

        for interface, ipv4_name in set(net_devices):
            lines.append(f'{ipv4_name}->SetUp({interface});')
        return lines

    def assign_device_ips(G : Graph):
        lines = []
        for node, data in G.nodes(data=True):
            if data['role'] not in EDeviceRole.configurables():
                continue
            lines.append('')
            lines.append(f'Ptr<NetDevice> {node}NetDevice= {node}->GetDevice(0);')
            lines.append(f'Ptr<Ipv4> ipv4proto{node} = {node}->GetObject<Ipv4>();')
            lines.append(f'uint32_t {node}interface = ipv4proto{node}->AddInterface({node}NetDevice);')
            ip, ipv4Mask = data['ip'], str(data['subnet']).split('/')[-1]
            lines.append(f'Ipv4InterfaceAddress ipv4Addr{node} = Ipv4InterfaceAddress(Ipv4Address("{ip}"), Ipv4Mask("/{ipv4Mask}"));')
            lines.append(f'ipv4proto{node}->AddAddress({node}interface, ipv4Addr{node});')
            lines.append(f'ipv4proto{node}->SetUp({node}interface);')
        return lines
    
    def install_udp(G : Graph, packet_map : dict):
        def add_servers(udp_nodes):
            lines = ['']
            lines.append(f'NodeContainer udpNodes({", ".join(udp_nodes)});')
            lines.append(f'ApplicationContainer serverApp = server.Install(udpNodes);')
            lines.append(f'serverApp.Start(Seconds(0.5));')
            lines.append(f'serverApp.Stop(Seconds(simDurationSeconds));')
            lines.append(f'serverApp.Stop(Seconds(simDurationSeconds));')
            return lines

        def add_sink(node, server_ip, server_index):
            lines = []
            lines.append('')
            prefix = f'{node}_{server_index}'
            lines.append(f'UdpClientHelper client{prefix}(Ipv4Address("{server_ip}"), udpEchoPort);')
            for key, value in [('MaxPackets', 'UintegerValue()'), ('Interval', 'TimeValue(interPacketInterval)'), ('PacketSize', 'UintegerValue(packetSize)')]:
                lines.append(f'client{prefix}.SetAttribute("{key}", {value});')
            lines.append(f'ApplicationContainer {prefix}ClientApp = client{prefix}.Install({node});')
            lines.append(f'{prefix}ClientApp.Start(Seconds(0.5));')
            lines.append(f'{prefix}ClientApp.Stop(Seconds(simDurationSeconds));')
            return lines

        udp_packets = [packet for _, packets in packet_map.items() for packet in packets if packet[EParameterKey.PROTOCOL.value] == 'udp']
        servers = set([packet[EParameterKey.DST.value] for packet in udp_packets])
        lines = add_servers(servers)
        for i, server in enumerate(servers):
            related_sinks = set([(packet[EParameterKey.SRC.value], packet[f'{EParameterKey.DST.name.lower()}_ip']) for packet in udp_packets if packet[EParameterKey.DST.value] == server])
            for node, server_ip in related_sinks:
                lines += add_sink(node, server_ip, i)
        
        return lines



    def install_tcp(G : Graph, packet_map):
        def add_server(node, ip, port, prefix):
            lines = ['']
            lines.append(f'Ptr<Socket> {prefix}ServerSocket = Socket::CreateSocket({node}, TcpSocketFactory::GetTypeId());')
            lines.append(f'InetSocketAddress {prefix}ServerAddress = InetSocketAddress(Ipv4Address("{ip}"), {port});')
            lines.append(f'{prefix}ServerSocket->Bind({prefix}ServerAddress);')
            lines.append(f'{prefix}ServerSocket->Listen();')
            lines.append(f'{prefix}ServerSocket->SetRecvCallback(MakeCallback(&ReceivePacket));')
            return lines

        def add_socket(node_id, server, prefix, message_length = 1024, send_interval = 0.5):
            lines = ['']
            lines.append(f'Ptr<Socket> {prefix}ClientSocket = Socket::CreateSocket({node_id}, TcpSocketFactory::GetTypeId());')
            lines.append(f'{prefix}ClientSocket->Connect({server}ServerAddress);')
            lines.append(f'Simulator::Schedule(Seconds(2.0), &SendPacket, {prefix}ClientSocket, {message_length}, Seconds({send_interval}));')
            return lines
        
        tcp_packets = [packet for _, packets in packet_map.items() for packet in packets if packet[EParameterKey.PROTOCOL.value] == 'tcp']

        servers = set([(packet[EParameterKey.DST.value], packet[EParameterKey.DST_PORT.value]) for packet in tcp_packets])
        lines = []
        for i, (server, port) in enumerate(servers):
            server_prefix = f'{server}_{port}'
            lines += add_server(server, G.nodes[server]['ip'], port, server_prefix)

            packets = [packet for packet in tcp_packets if packet[EParameterKey.DST.value] == server]
            for j, packet in enumerate(packets):
                node = packet[EParameterKey.SRC.value]
                prefix = f'{node}_{i}_{j}'
                send_interval = round(1/ packet['packets_per_second'], 3)
                message_length = packet['packet_size']
                lines += add_socket(node, server_prefix, prefix, message_length, send_interval)

        return lines

    with open(load_resource('templates', 'template.cc'), 'r') as resource_file, open(join(location, NS3_FOLDER, 'genesis.cc'), 'w') as output_file:
        for line in resource_file.readlines():
            output_file.write(line)
            for func in [create_nodes, add_names, create_links, create_bridges, install_bridges, install_end_device_ip_stack, install_router_ip_stack, assign_router_ips, assign_device_ips]:
                if func.__name__ in line:
                    output_file.writelines(map(lambda x: f'\t{x}\n', func(G)))

            for func in [install_udp, install_tcp]:
                if func.__name__ in line:
                    output_file.writelines(map(lambda x: f'\t{x}\n', func(G, generated_packets)))

def to_image(G, location, seed=None):
    g_copy = G.copy()
    color_map, size_map, border_map, shape_map = [], [], [], []
    for n in g_copy.nodes:
        role = EDeviceRole.from_device_id(n)

        if role == EDeviceRole.PORT:
            color_map.append('lightgray')
            size_map.append(50)
        elif role == EDeviceRole.ROUTER:
            color_map.append('black')
            border_map.append('black')
            size_map.append(200)
        elif role == EDeviceRole.SWITCH:
            color_map.append('white')
            border_map.append('black')
            size_map.append(150)
        elif role == EDeviceRole.CONTROLLER:
            color_map.append('#F8CECC')
            border_map.append('#B85450')
            shape_map.append('d')
            size_map.append(100)
        elif role == EDeviceRole.SERVER:
            color_map.append('#DAE8FC')
            border_map.append('#6C8EBF')
            shape_map.append('o')
            size_map.append(100)
        else:
            border_map.append('#D6B656')
            color_map.append('#FFF2CC')
            shape_map.append('o')
            size_map.append(100)

    draw(g_copy, node_color=color_map, node_size=size_map, edgecolors=border_map, with_labels=False, pos=spring_layout(g_copy, seed=seed))
    savefig(join(location, f'graph.png'), dpi=300)
    savefig(join(location, f'graph.jpg'), dpi=300)
    close()



from datetime import datetime
from itertools import product
from os.path import join
from hses_genesis.utils.enum_objects import EPacketDecision

def to_save_file(location, node, ruleset, default_action = EPacketDecision.DROP):
    with open(join(location, f'{node.lower()}-iptables-save'), 'w') as file:
        file.write(f'# Generated on {datetime.now().strftime("%a %b %d %H:%M:%S %Y")}\n')
        file.write('*filter\n')
        [file.write(f':{chain} {default_action.name} [0:0]\n') for chain in ['INPUT', 'FORWARD', 'OUTPUT']]
        for rule in ruleset:
            file.write(f'-A INPUT {rule}\n')
        file.write('COMMIT\n')
        file.write(f'# Completed on {datetime.now().strftime("%a %b %d %H:%M:%S %Y")}')

def to_zimpl_parsable(location, rulesets):
    with open(join(location, 'ruleset.txt'), 'w') as file:
        header_fields = [f'{prefix}_{suffix}' for prefix, suffix in product(['src', 'dst', 'prot', 'src_port', 'dst_port', 'state'], ['start', 'end'])] + ['action']
        for i, header in enumerate(header_fields):
            file.write(f'# <router,x,{i}>{header}\n')

        for router, ruleset in rulesets:
            for i, rule in enumerate(ruleset):
                for j, v in enumerate(rule):
                    file.write(f'{router},{i},{j},{v}\n')


from random import Random
from networkx import Graph, has_path, shortest_simple_paths
from hses_genesis.utils.enum_objects import EDeviceRole, EService, EState, ETrafficProfile, ENetworkLayer
from hses_genesis.utils.functions import print_information

def print_connection_info(total_connections, forbidden_connections, intrasubnet_connections, intersubnet_connections, sampled_connections, traffic_profile = ETrafficProfile.STRICT_ISOLATION):
    print_information('CONNECTION INFORMATION', {
        'number of possible device connections' : len(total_connections),
        '- of which forbidden': f'{len(forbidden_connections)}, i.e., violates {traffic_profile.name.lower().replace("_", " ")} profile',
        '- of which intrasubnet' : len(intrasubnet_connections),
        '- of which intersubnet' : len(intersubnet_connections),
        '-- of which sampled' : len(sampled_connections)
    })

class CommunicationGenerator():

    def __init__(self, seed) -> None:
        self.random = Random(seed)

    def get_intrasubnet_connections(self, G : Graph):
        connections = []
        for (src_node, src_data) in G.nodes(data=True):
            if src_data['role'] not in EDeviceRole.configurables():
                continue

            initiating_services = [(service_name, service_states) for (service_name, service_states) in src_data['services'] if EState.NEW in service_states] 
            if not initiating_services:
                continue

            for (dst_node, dst_data) in G.nodes(data=True):
                if (src_node == dst_node) or (src_data['subnet'] != dst_data['subnet']) or (dst_data['role'] not in EDeviceRole.configurables()):
                    continue

                overlapping_services = [src_service for (src_service, _) in src_data['services'] if src_service in [dst_service for (dst_service, dst_service_states) in dst_data['services'] if (EState.ESTABLISHED in dst_service_states or EState.RELATED in dst_service_states)]]
                
                if not overlapping_services:
                    continue

                connections.extend([(src_node, dst_node, service) for service in overlapping_services])
        return connections
                
    def get_connections(self, G : Graph, traffic_profile = ETrafficProfile.DISTRIBUTED_CONTROL, upper_connection_bound = -1):
        def of_same_branch(a_branch, b_branch):
            return a_branch == b_branch or a_branch.startswith(b_branch) or b_branch.startswith(a_branch)
        
        def is_central_server(data):
            return data['role'] == EDeviceRole.SERVER and data['layer'] == ENetworkLayer.CONNECTIVITY.name
        
        def is_allowed_in_strict_isolation(src_data, dst_data):
            """
            Returns True if source and target are controllers within the same branch.
            """
            return of_same_branch(src_data['branch'], dst_data['branch']) and all(role == EDeviceRole.CONTROLLER for role in [src_data['role'], dst_data['role']])
        
        def is_allowed_in_converged_networks(src_data, dst_data):
            """
            Returns True if source or target is a server located in enterprise layer and the other communication partner is a controller anywhere in the network.
            """
            return EDeviceRole.CONTROLLER in [src_data['role'], dst_data['role']] and any(is_central_server(data) for data in [src_data, dst_data])
        
        def is_allowed_in_distributed_control(src_data, dst_data):
            """
            Returns True if both source and target are controllers or server to any device
            """
            return all(role == EDeviceRole.CONTROLLER for role in [src_data['role'], dst_data['role']]) or any(is_central_server(data) for data in [src_data, dst_data])

        total_connections = []
        forbidden_connections = []
        intrasubnet_connections = []
        intersubnet_connections = []

        for (src_node, src_data) in G.nodes(data=True):
            if src_data['role'] not in EDeviceRole.configurables():
                continue
            
            initiating_services = [service for service in src_data['services'] if EState.NEW in src_data['role'].possible_service_states(service)] 
            if not initiating_services:
                continue

            for (dst_node, dst_data) in G.nodes(data=True):
                if (src_node == dst_node) or (dst_data['role'] not in EDeviceRole.configurables()):
                    continue

                overlapping_services = [src_service for src_service in src_data['services'] if src_service in dst_data['services']]
                if not overlapping_services or not has_path(G, src_node, dst_node):
                    continue

                connections = [(src_node, dst_node, service) for service in overlapping_services]
                total_connections.extend(connections)

                is_intrasubnet = src_data['subnet'] == dst_data['subnet']
                if is_intrasubnet:
                    intrasubnet_connections.extend(connections)
                    continue

                if is_allowed_in_strict_isolation(src_data, dst_data):
                    intersubnet_connections.extend(connections)
                    continue
                elif traffic_profile == ETrafficProfile.STRICT_ISOLATION:
                    forbidden_connections.extend(connections)
                    continue

                if is_allowed_in_converged_networks(src_data, dst_data):
                    intersubnet_connections.extend(connections)
                    continue
                elif traffic_profile == ETrafficProfile.CONVERGED_NETWORKS:
                    forbidden_connections.extend(connections)
                    continue

                if is_allowed_in_distributed_control(src_data, dst_data):
                    intersubnet_connections.extend(connections)
                    continue

                forbidden_connections.extend(connections)

        if upper_connection_bound < 0 or len(intersubnet_connections) < upper_connection_bound:
            sampled_connections = intersubnet_connections
        else:
            sampled_connections = self.random.sample(intersubnet_connections, k=upper_connection_bound)

        print_connection_info(total_connections=total_connections, forbidden_connections=forbidden_connections, intrasubnet_connections=intrasubnet_connections, intersubnet_connections=intersubnet_connections, sampled_connections=sampled_connections, traffic_profile=traffic_profile)

        return intrasubnet_connections, sampled_connections
    
    def get_affected_connections(self, G : Graph, router : str, allowed_connections):
        output = []
        for source, target, p, sport, dport in allowed_connections:
            if any(router in path for path in shortest_simple_paths(G, source, target)):
                output.append((G.nodes[source]['ip'], G.nodes[target]['ip'], p, sport, dport))
        return output
    
    def generate_packet(self, service : EService, src_node : str, src_data : dict, dst_node : str, dst_data : dict, protocol : str, port : int, is_high_sender = False):
        return {
            'SourceName' : src_node,
            'DestinationName' : dst_node,
            
            # Ethernet
            'SourceMac' : src_data['mac'],
            'DestinationMac' : dst_data['mac'],
            'EtherType' : 'IPv4',

            # 802.1Q
            'PriorityCodePoint' : service.value.priority,
            'DropEligibleIndicator' : service.value.dei, 
            'VlanIndicator' : 1,
            
            # IP
            'Protocol' : protocol,
            'SourceIp' : src_data['ip'],
            'DestinationIp' : dst_data['ip'],

            # TCP/UDP
            'SourcePort' : port,
            'DestinationPort' : port,

            # General
            'PacketSize' : self.random.randint(service.value.packet_size_range[0], service.value.packet_size_range[1]),
            'PacketsPerSecond' : self.random.randint(10, 100) * (2 if is_high_sender else 1)
        }


from ipaddress import ip_address, ip_network
from itertools import product
from re import match, search
from hses_genesis.utils.constants import ACTION_PATTERN, CHAIN_PATTERN, FULL_RANGES, PROTOCOLS, RULE_CONTAINER_PATTERN, SUBNET_INDICATOR, TABLE_PATTERN, RULE_VALUE_SEPARATOR, WILDCARD, MEDIUM_SEPARATOR
from hses_genesis.utils.enum_objects import EParameterKey, EParameterType, EPacketDecision

def __unify_value(value : str, parameter_key : EParameterKey):
    parameter_type = EParameterType.from_parameter_key(parameter_key)
    
    if value == WILDCARD:
        return FULL_RANGES[parameter_type]
    
    if parameter_type == EParameterType.IP:
        return int(ip_address(value))
    
    if parameter_key == EParameterKey.PROTOCOL:
        for acronym, protocol_number in PROTOCOLS.items():
            if value == protocol_number or str(value).lower() == acronym:
                return PROTOCOLS[value.lower()]
    return int(value)

def __get_unified_ranges(value : str, parameter_key : EParameterKey):
    if RULE_VALUE_SEPARATOR in value:
        return [(__unify_value(v, parameter_key), __unify_value(v, parameter_key)) for v in value.split(RULE_VALUE_SEPARATOR)]
    elif MEDIUM_SEPARATOR in value:
        start, end = value.split(MEDIUM_SEPARATOR)
        return [(__unify_value(start, parameter_key), __unify_value(end, parameter_key))]
    elif SUBNET_INDICATOR in value:
        network_address = ip_network(value, strict=False)
        start, end = network_address[0], network_address[-1]
        return [(__unify_value(start, parameter_key), __unify_value(end, parameter_key))]
    
    start = __unify_value(value, parameter_key)
    if isinstance(start, tuple):
        return [start]
    return [(start, start)]

def negate_value_ranges(value_ranges : list[tuple[int, int]], parameter_type : EParameterType):
    start, end = FULL_RANGES[parameter_type]
    ranges = [(start, start)] + value_ranges + [(end, end)]
    negated = []
    for i in range(len(ranges) - 1):
        (_, front_end), (r_s, _) = ranges[i], ranges[i+1]
        new_range = (front_end + (1 if i > 0 else 0), r_s - (1 if i < len(ranges) - 2 else 0))
        if new_range not in negated:
            negated.append(new_range)
    return negated

def parse_parameters(value : str):
    parameter_values = []
    for parameter_key in EParameterKey:
        parameter_match = search(f'(?P<negated>!\s)?-{parameter_key.value}\s(?P<value>[^\s]+)', value)
        if parameter_match:
            groups = parameter_match.groupdict()
            parameter_value : str = groups['value']
            ranges = __get_unified_ranges(parameter_value, parameter_key)
            if 'negated' in groups.keys() and groups['negated'] != None:
                ranges = negate_value_ranges(ranges, EParameterType.from_parameter_key(parameter_key))
            parameter_values.append(ranges)
        else:
            parameter_type = EParameterType.from_parameter_key(parameter_key)
            parameter_values.append([FULL_RANGES[parameter_type]])
    return parameter_values


def parse_rule(value : str, from_node = False):
    rule_pattern = (RULE_CONTAINER_PATTERN + '.+' + ACTION_PATTERN) if not from_node else ('^.+' + ACTION_PATTERN)
    general_match = match(rule_pattern, value)
    if not general_match:
        return None

    chain = general_match.group('chain') if 'chain' in general_match.groupdict().keys() else 'INPUT'
    action = EPacketDecision.from_str(general_match.group('action'))
    
    rules = [(chain, parameters, action) for parameters in product(*parse_parameters(value))]
    return rules

def parse_tables(lines : list[str], from_node = False, debug = False):
    tables = []
    if from_node:
        tables.append((
            'filter',
            [('INPUT', 'DROP'), ('FORWARD', 'DROP'), ('OUTPUT', 'DROP')],
            []
        ))
    for line in lines:
        line_value = line.strip()
        hit = match(TABLE_PATTERN, line_value)
        if hit:
            tables.append((hit.group('table'), [], []))
            continue

        hit = match(CHAIN_PATTERN, line_value)
        if hit:
            tables[-1][1].append((hit.group('chain'), EPacketDecision.from_str(hit.group('default_action'))))
            continue
        
        rule = parse_rule(line_value, from_node)
        if rule:
            tables[-1][2].extend(rule)
            continue
        
        if debug:
            print(f'Skipped parsing of line: {line_value}')
    return tables


from math import ceil, log2
from random import Random
from networkx import Graph, compose, connected_components, cycle_graph, path_graph, neighbors, star_graph, descendants_at_distance
from hses_genesis.utils.enum_objects import EDeviceRole, ENetworkLayer, EPacketDecision, EService, ESubnetTopologyStructure
from hses_genesis.parsing.configuration import LayerDefinition
from ipaddress import ip_network, IPv4Network

def get_graph_information(G : Graph):
    return {
        'number of nodes' : len(G.nodes),
        '- of which ports' : len([n for n in G.nodes if "#" in n]),
        '- of which routers' : len([n for n in G.nodes if "#" not in n and EDeviceRole.from_device_id(n) == EDeviceRole.ROUTER]),
        '- of which switches' : len([n for n in G.nodes if "#" not in n and EDeviceRole.from_device_id(n) == EDeviceRole.SWITCH]),
        '- of which end devices' : len([n for n in G.nodes if "#" not in n and EDeviceRole.from_device_id(n) not in [EDeviceRole.SWITCH, EDeviceRole.ROUTER]]),
        'number of edges' : len(G.edges),
        '- of which cables' : len([(s, d) for (s, d) in G.edges if s not in d and d not in s]),
        'number of subnets' : len(list(set([data["subnet"] for _, data in G.nodes(data=True) if data['role'] != EDeviceRole.ROUTER]))),
        'number of subgraphs' : f'{len(list(connected_components(G)))} (should always be 1)'
    }

def calculate_subnets(layer_definitions : list[LayerDefinition]):
    required_number_of_ips = []
    for i, layer_definition in enumerate(layer_definitions):
        switch_count = layer_definition.switch_count.CURRENT
        if switch_count < 0 and i < len(layer_definitions) - 1:
            switch_count = (2 * layer_definition.subnet_descendants.CURRENT) + (1 if layer_definition.structure_distribution.get(ESubnetTopologyStructure.STAR, 0) > 0 else 0)
        ip_count = switch_count + (switch_count * layer_definition.max_hosts_per_switch.CURRENT) + 4 # +4 due to worst case number of routers in a subnet + number of additional ips between routers
        
        if i < len(layer_definitions) - 1:
            child_count = layer_definition.subnet_descendants.CURRENT if i < len(layer_definitions) - 1 else 0
            ip_count += 2 * child_count
        required_number_of_ips.append(ip_count)
    worst_case_subnet_size = max(required_number_of_ips) + 2 # +2 due to diff between total number of hosts and usable number of hosts
    netmask_size = max(16, min(32 - ceil(log2(worst_case_subnet_size)), 32))
    base_network = ip_network(f'192.0.0.0/16')
    return list(base_network.subnets(new_prefix=netmask_size))

class TopologyGenerator():
    def __init__(self, seed : int) -> None:
        self.existing_devices = []
        self.existing_subnets = []
        self.existing_macs = []
        self.random = Random(seed)

    def generate_device(self, device_role : EDeviceRole, initial_index = 0):
        index = initial_index
        while f'{device_role.value}{index}' in self.existing_devices:
            index += 1
        self.existing_devices.append(f'{device_role.value}{index}')
        return f'{device_role.value}{index}'

    def assign_node_informations(self, G : Graph, layer : ENetworkLayer, branch_id : str, subnet_ip : IPv4Network):
        if len([n for n in G.nodes if EDeviceRole.from_device_id(n) not in [EDeviceRole.PORT]]) > 253:
            raise Exception(f'Too many devices in a single subnet! Only {len(subnet_ip.hosts)} devices are supported.')

        self.existing_subnets.append(subnet_ip)
        ips = list(subnet_ip.hosts())
        
        for device in G.nodes:
            mac_index = len(self.existing_macs)
            mac_address = "02:00:00:%02x:%02x:%02x" % ((mac_index >> 16) & 0xFF, (mac_index >> 8) & 0xFF, mac_index & 0xFF)
            G.nodes[device]['mac'] = mac_address
            self.existing_macs.append(mac_address)
            
            device_role = EDeviceRole.from_device_id(device)
            G.nodes[device]['layer'] = layer.name
            G.nodes[device]['branch'] = branch_id
            G.nodes[device]['role'] = device_role
            possible_services = device_role.possible_services()
            services = self.random.sample(possible_services, k=self.random.randint(1,len(possible_services)))
            G.nodes[device]['services'] = services
            ip = str(ips.pop(0))
            if device_role == EDeviceRole.ROUTER:
                G.nodes[device]['subnet'] = [str(subnet_ip)]
                G.nodes[device]['ip'] = [str(ip)]
                G.nodes[device]['default_action'] = EPacketDecision.DROP
                G.nodes[device]['ruleset'] = []
            else:
                G.nodes[device]['subnet'] = str(subnet_ip)
                G.nodes[device]['ip'] = str(ip)
        return G, ips
    
    def get_topology_type(self, topology_distribution : dict[ESubnetTopologyStructure, int]):
        tmp_distribution : dict[ESubnetTopologyStructure, int] = topology_distribution.copy()
        if all(percentage == 0 for percentage in tmp_distribution.values()):
            tmp_distribution = {key : 1 for key, _ in tmp_distribution.items()}

        return self.random.choices(list(tmp_distribution.keys()), tmp_distribution.values(), k = 1)[0]

    def get_barebone_subnet(self,
                            number_of_switches : int,
                            subnet_structure : ESubnetTopologyStructure,
                            has_parent = False):

        switches = [self.generate_device(EDeviceRole.SWITCH, index) for index in range(number_of_switches)]
        routers = []
        if has_parent:
            routers = [self.generate_device(EDeviceRole.ROUTER) for _ in range(2 if subnet_structure in ESubnetTopologyStructure.redundant_types() else 1)]

        forwarding_devices = switches + routers

        if subnet_structure == ESubnetTopologyStructure.LINE:
            G = path_graph(forwarding_devices)
        elif subnet_structure == ESubnetTopologyStructure.RING:
            G = cycle_graph(forwarding_devices)
        elif subnet_structure == ESubnetTopologyStructure.STAR:
            G = star_graph(forwarding_devices)
        else:
            G = Graph([(s, e) for s in forwarding_devices for e in forwarding_devices if s != e])
        return G

    def add_devices(self, G : Graph, max_hosts_per_owner : int, host_types : dict[EDeviceRole, int], subnet_structure : ESubnetTopologyStructure):
        def is_center_node(G, node):
            return sum([1 for n in neighbors(G, node) if EDeviceRole.from_device_id(n) == EDeviceRole.SWITCH]) > 2
        
        if sum([max(0, v) for v in host_types.values()]) == 0 and sum([min(0, v) for v in host_types.values()]) == 0:
            return G

        device_owners = [node for node in G.nodes() if EDeviceRole.from_device_id(node) == EDeviceRole.SWITCH and not (subnet_structure == ESubnetTopologyStructure.STAR and is_center_node(G, node))]
        if len(device_owners) == 0:
            raise Exception('InvalidConfigurationException: Layer instance without switches generated. Most likely your configuration file does not specify a switch_count in the last layer_definition.')
        
        if len(device_owners) * max_hosts_per_owner < sum([max(value, 0) for value in host_types.values()]):
            raise Exception('InvalidConfigurationException: Your configuration of fixed devices exceeds the available space.')
        

        for device_type, device_count in host_types.items():
            if device_count <= 0:
                continue

            if len(device_owners) - device_count <= 0:
                chosen_owners = device_owners.copy()
            else:
                chosen_owners = self.random.sample(device_owners, k = device_count)

            for chosen_owner in chosen_owners:
                device = self.generate_device(device_type)
                G.add_node(device)
                G.add_edge(chosen_owner, device)

        filling_hosts = {host_type : abs(count) for host_type, count in host_types.items() if count < 0}
        
        if not filling_hosts:
            return G

        for switch in device_owners:
            connected_devices = sum([1 for n in neighbors(G, switch) if EDeviceRole.from_device_id(n) in EDeviceRole.configurables()])
            remaining_space = max_hosts_per_owner - connected_devices

            if remaining_space <= 0:
                continue

            for _ in range(remaining_space):
                device_type = self.random.choices(list(filling_hosts.keys()), weights=list(filling_hosts.values()), k=1)[0]
                device = self.generate_device(device_type)
                G.add_node(device)
                G.add_edge(switch, device)
        return G
        
    def generate_subnet(self,
                        subnet_structure : ESubnetTopologyStructure,
                        switch_count : int,
                        hosts_per_owner : int,
                        host_types : list[EDeviceRole],
                        has_parent = False) -> Graph:
        
        G = self.get_barebone_subnet(switch_count, subnet_structure, has_parent) 

        G = self.add_devices(G, max_hosts_per_owner=hosts_per_owner,
                             host_types=host_types, subnet_structure=subnet_structure)

        return G


    def generate_network(self, layer_definitions : list[LayerDefinition], available_subnet_ips : list, layer_index = 0, branch_id = '0'):
        """
        The amount of switches per subnet/layer is dependant on its specified dimension as well as 2 * the number of child/lower layers.
        The additional switches are necessary, since every router has to have a connectable switch in an upper layer.
        """
        layer_definition = layer_definitions[layer_index]

        switch_count = layer_definition.switch_count.CURRENT
        if switch_count < 0 and layer_index < len(layer_definitions) - 1:
            switch_count = (1 if layer_definition.structure_distribution.get(ESubnetTopologyStructure.STAR, 0) > 0 else 0) + (layer_definition.subnet_descendants.CURRENT * 2)

        network = self.generate_subnet(
            subnet_structure=self.get_topology_type(layer_definition.structure_distribution),
            switch_count=switch_count,
            hosts_per_owner=layer_definition.max_hosts_per_switch.CURRENT,
            host_types=layer_definition.host_types,
            has_parent=layer_definitions.index(layer_definition) > 0
        )

        subnet_ip = available_subnet_ips.pop(0)
        network, open_subnet_ips = self.assign_node_informations(network, layer_definition.layer_classification, branch_id=branch_id, subnet_ip=subnet_ip)

        if layer_index >= len(layer_definitions) - 1:
            return network
        
        switches = [node for node, data in network.nodes(data=True) if data['role'] == EDeviceRole.SWITCH]
        self.random.shuffle(switches)

        children = [self.generate_network(layer_definitions, available_subnet_ips, layer_index=layer_index+1, branch_id=f'{branch_id}.{i}') for i in range(layer_definition.subnet_descendants.CURRENT)]

        for child_index, child_network in enumerate(children):
            network : Graph = compose(network, child_network)

            child_branch = f'{branch_id}.{child_index}'
            child_routers = [node for node, data in child_network.nodes(data=True) if data['role'] == EDeviceRole.ROUTER and data['branch'] == child_branch]
            child_subnet = set(data['subnet'] for _, data in child_network.nodes(data=True) if data['role'] != EDeviceRole.ROUTER and data['branch'] == child_branch)
            child_subnet = list(child_subnet)[0]
            for router_index, router in enumerate(child_routers):
                network.nodes[router]['ip'].append(str(open_subnet_ips.pop(0)))
                network.nodes[router]['subnet'].append(str(subnet_ip))

                switch_index = ((len(child_routers) * child_index) + router_index) % len(switches)
                network.add_edge(router, switches[switch_index])

        return network


from math import floor
from hses_genesis.utils.constants import TERMINAL_WIDTH
from os.path import join, dirname, exists


def print_information(title : str, key_values : dict):
    prefix_len = int(floor(TERMINAL_WIDTH - len(title) - 2) / 2)
    print('-' * prefix_len, title, '-' * (TERMINAL_WIDTH - (prefix_len + len(title) + 2)))
    padding = max([len(key) for key in key_values.keys()]) + 2
    for key, value in key_values.items():
        print(str(key) + ':' + (padding - len(key)) * ' ' + str(value))
    print('-' * TERMINAL_WIDTH)

def load_resource(folder, file):
    file_location = join(dirname(dirname(__file__)), 'resources', folder, file)
    if exists(file_location):
        return file_location
    else:
        raise Exception(f'Resource does not exist: {file_location}')


from datetime import datetime
from hses_genesis.parsing.configuration import GenerationConfig
from os.path import join, exists
from os import makedirs

from hses_genesis.utils.constants import GRAPH_FOLDER, NS3_FOLDER, OMNET_FOLDER, PACKET_FOLDER, RULESET_FOLDER, ZIMPL_FOLDER

def setup_base_location(config : GenerationConfig, output_location, config_name = 'config'):
    testrun_id = datetime.now().strftime('%y-%m-%d-%H-%M-%S')
    location = join(output_location, config_name, testrun_id)
    makedirs(location, exist_ok = True)

    config.to_total_file(join(location, 'config.json'))

    with open(join(location, '.genesistag'), 'x') as file:
        file.write(config.to_total_str())

    return location

def setup_run_location(config : GenerationConfig, base_location : str, run_label, export_iptables_files = False, export_omnet_files = False, export_ns3_files = False, export_zimpl_parsables = False):
    i = 0
    while True:
        run_location = join(base_location, f'{run_label}-{i}')
        if not exists(run_location):
            break
        i += 1
    makedirs(run_location, exist_ok=True)

    config.to_run_file(join(run_location, 'config.json'))
    with open(join(run_location, '.genesistag'), 'x') as file:
        file.write(config.to_run_str())

    for subfolder in [GRAPH_FOLDER, PACKET_FOLDER] + ([RULESET_FOLDER] if export_iptables_files else []):
        sublocation = join(run_location, subfolder)
        makedirs(sublocation, exist_ok=True)

    if export_omnet_files:
        makedirs(join(run_location, OMNET_FOLDER), exist_ok=True)

    if export_ns3_files:
        makedirs(join(run_location, NS3_FOLDER), exist_ok=True)

    if export_zimpl_parsables:
        makedirs(join(run_location, ZIMPL_FOLDER), exist_ok=True)

    return run_location


from collections import OrderedDict
from ipaddress import ip_address
from random import Random
from hses_genesis.utils.constants import FULL_RANGES, PROTOCOLS, WILDCARD
from hses_genesis.utils.enum_objects import EPacketDecision, EParameterType, EState

class NetworkConfigurationGenerator():
    def __init__(self, seed) -> None:
        self.seed = seed

    @staticmethod
    def rule_to_str(raw_rule):
        src, dst, (prot_start, prot_end), sport, dport, (state_start, state_end), action = raw_rule
        rule = []
        for (start, end) in [src, dst]:
            if start == end:
                rule.append(str(ip_address(start)))
                continue

            start_ip, end_ip = ip_address(start), ip_address(end)
            if FULL_RANGES[EParameterType.PROTOCOL] == (int(start_ip), int(end_ip)):
                rule.append(WILDCARD)
                continue

            rule.append(f'{start_ip}:{end_ip}')

        if prot_start != prot_end:
            raise Exception('Multi protocol rules not implemented yet...')
        
        if isinstance(prot_start, int):
            rule.append({v: k for (k, v) in PROTOCOLS.items()}[prot_start])
        else:
            rule.append(prot_start.lower())

        for (start, end) in [sport, dport]:
            if start == end:
                rule.append(str(start))
                continue
            if FULL_RANGES[EParameterType.NUMBER] == (start, end):
                rule.append(WILDCARD)
                continue
            rule.append(f'{start}:{end}')

        if state_start != EState.NONE:
            rule.append(','.join([EState.from_value(i).name for i in range(state_start.value, state_end.value + 1)]))
        else:
            rule.append('')

        src, dst, prot, sport, dport, states = rule
        return f'-s {src} -d {dst} -p {prot} --sport {sport if sport != 0 else WILDCARD} --dport {dport if dport != 0 else WILDCARD}' + (f' -m state --state {states}' if states else '') + f' -j {action.name}'

    @staticmethod
    def to_numerical_representation(raw_rule):
        ip_addresses = [int(ip_address(value)) for value_range in raw_rule[0:2] for value in list(value_range)]
        protocols = [PROTOCOLS[value.lower()] if isinstance(value, str) else value for value in raw_rule[2]]
        ports = [int(value) for value_range in raw_rule[3:5] for value in list(value_range)]
        states = [state.value for state in list(raw_rule[5])]
        action = [raw_rule[6].value]
        return ip_addresses + protocols + ports + states + action

    def generate_ruleset(self, communication_relations, anomaly_count : int, stateful_percentage : int):
        ruleset = []
        truth_weight = stateful_percentage / 100 if stateful_percentage != 0 else 0
        for i, (src, dst, prot, sport, dport) in enumerate(communication_relations):
            is_stateful_rule = Random(self.seed + i).choices([True, False], weights = [truth_weight, 1 - truth_weight], k = 1)[0]
            if is_stateful_rule:
                rule = (src,src), (dst,dst), (prot,prot), FULL_RANGES[EParameterType.NUMBER], (dport,dport), (EState.NEW, EState.RELATED), EPacketDecision.ACCEPT
                mirror = (dst,dst), (src,src), (prot,prot), (sport,sport), FULL_RANGES[EParameterType.NUMBER], (EState.ESTABLISHED, EState.RELATED), EPacketDecision.ACCEPT
            else:
                rule = (src,src), (dst,dst), (prot,prot), FULL_RANGES[EParameterType.NUMBER], (dport,dport), (EState.NONE, EState.NONE), EPacketDecision.ACCEPT
                mirror = (dst,dst), (src,src), (prot,prot), (sport,sport), FULL_RANGES[EParameterType.NUMBER], (EState.NONE, EState.NONE), EPacketDecision.ACCEPT
            ruleset.extend([rule, mirror])

        if anomaly_count == 0:
            return ruleset

        src_matching_rule_groups : dict[str, list] = {}
        dst_matching_rule_groups : dict[str, list] = {}
        for i, ((src, _), (dst, _), _, _, _, _, _) in enumerate(ruleset):
            for tag_value, container in [src, src_matching_rule_groups], [dst, dst_matching_rule_groups]:
                if tag_value not in container:
                    container[tag_value] = [ruleset[i]]
                else:
                    container[tag_value].append(ruleset[i])

        counter = 0
        for ranged_value_index, container in [[1, src_matching_rule_groups], [0, dst_matching_rule_groups]]:
            if counter >= anomaly_count:
                break

            for key in container.keys():
                if counter >= anomaly_count:
                    break

                if len(container[key]) > 1:
                    rules = container[key]
                    sorted(rules, key=lambda x: x[ranged_value_index])
                    
                    remaining_anomalies = anomaly_count - counter
                    if len(rules) > remaining_anomalies:
                        rules = rules[:remaining_anomalies]
                    counter += len(rules)

                    src, dst, prot, sport, dport = [(f'{rules[0][i]}:{rules[-1][i]}' if (rules[0][i] != rules[-1][i]) else rules[0][i]) if i < 2 else WILDCARD for i, _ in enumerate(['s', 'd', 'p', 'sport', 'dport'])]

                    ruleset.insert(ruleset.index(rules[-1]) + 1, ((src, src), (dst, dst), (prot, prot), (sport, sport), (dport, dport), (EState.NONE, EState.NONE), EPacketDecision.DROP))

        if counter < anomaly_count:
            for ranged_value_index, container in [[1, src_matching_rule_groups], [0, dst_matching_rule_groups]]:
                if counter >= anomaly_count:
                    break

                for key in container.keys():
                    if counter >= anomaly_count:
                        break

                    if len(container[key]) == 1:
                        src, dst, prot, sport, dport = [container[key][0][i] if ((i < 2) and (i != ranged_value_index)) else WILDCARD for i, _ in enumerate(['s', 'd', 'p', 'sport', 'dport'])]
                        ruleset.insert(ruleset.index(container[key][0]) + 1, (src, dst, prot, sport, dport, [], EPacketDecision.DROP))
                        counter += 1

        return list(OrderedDict.fromkeys([f'-s {src} -d {dst} -p {prot} --sport {sport if sport != 0 else WILDCARD} --dport {dport if dport != 0 else WILDCARD}' + f' -m state --state {",".join(list([s.name for s in states]))}' if len(states) > 0 else '' + f' -j {action.name}' for src, dst, prot, sport, dport, states, action in ruleset])), ruleset


from hses_genesis.utils.enum_objects import EParameterKey, EParameterType

TOPOLOGY_KEY = 'topology'
COMMUNICATION_KEY = 'communication'
SECURITY_KEY = 'security'

ITERATIONS_KEY = 'iterations'

LAYER_DEFINITIONS_KEY = 'layer_definitions'

SWITCH_COUNT_KEY = 'switch_count'
MAX_HOSTS_PER_SWITCH_KEY = 'max_hosts_per_switch'
STRUCTURE_KEY = "structure"
HOST_TYPES_KEY = "host_types"
REPETITIONS_KEY = 'repetitions'
SUBNET_DESCENDANTS_KEY = 'subnet_descendants'

TRAFFIC_PROFILE_KEY = "traffic_profile"
UPPER_CONNECTION_BOUND_KEY = "upper_connection_bound"

ANOMALY_COUNT_KEY = "anomaly_count"
STATEFUL_PERCENTAGE_KEY = "stateful_percentage"

SMALL_SEPARATOR = '·'

MEDIUM_SEPARATOR = ':'

BIG_SEPARATOR = '::'

UNSET_INDICATOR = '-'

ITERATION_SEPARATOR = '.'

TAG_ENCAPSULATOR = '#'

GENESIS_PACKAGE_NAME = 'hses_genesis'

RULE_CONTAINER_PATTERN = '^-A\s(?P<chain>[^\s]+)'

ACTION_PATTERN = '-j\s(?P<action>[^\s]+)'

TABLE_PATTERN = '^\*(?P<table>[^\s]+)$'

CHAIN_PATTERN = '^\:(?P<chain>[^\s]+)\s(?P<default_action>[^\s]+)'

RULE_VALUE_SEPARATOR = ','

WILDCARD = '*'

SUBNET_INDICATOR = '/'

GRAPH_FOLDER = 'graphs'

RULESET_FOLDER = 'rulesets'

OMNET_FOLDER = 'omnetpp'

NS3_FOLDER = 'ns3'

ZIMPL_FOLDER = 'zimpl'

PACKET_FOLDER = 'packets'

TERMINAL_WIDTH = 44

PROTOCOLS = {
    'ip' : 0,
    'tcp' : 6,
    'udp' : 17
}

def update_protocols(protocols : dict[str: int]):
    global PROTOCOLS
    PROTOCOLS = protocols

FULL_RANGES : dict[EParameterType, tuple[int, int]] = {
    EParameterType.IP: (1, 4294967295),
    EParameterType.NUMBER: (0, 65535),
    EParameterType.PROTOCOL: (0,255)
}

DEFAULT_TUPLE = [
    EParameterKey.SRC,
    EParameterKey.DST,
    EParameterKey.PROTOCOL,
    EParameterKey.SRC_PORT,
    EParameterKey.DST_PORT
]

PACKET_HEADERS = ['SourceName', 'DestinationName', 'SourceMac', 'DestinationMac', 'EtherType', 'PriorityCodePoint', 'DropEligibleIndicator', 'VlanIndicator', 'Protocol', 'SourceIp', 'DestinationIp', 'SourcePort', 'DestinationPort', 'PacketSize', 'PacketsPerSecond']